#!/usr/bin/env python
'''Compilation of observed and derived astrophysical trends for galaxies.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import math
import numpy as np
import os
import scipy.interpolate as interp
from scipy.optimize import newton
import sys

import galaxy_dive.utils.constants as constants
import galaxy_dive.utils.utilities as utilities

########################################################################

def circular_velocity( r_vir, m_vir ):
  '''Calculate the circular velocity of a halo in km/s.

  Args:
    r_vir (float or array-like) : The virial radius in pkpc.
    m_vir (float or array-like) : The halo mass in Msun.

  Returns:
    v_c : Circular velocity of the halo in km/s, indexed the same way that ahf_reader.mtree_halos[i] is.
  '''
  
  # Convert r_vir and m_vir to cgs
  r_vir_cgs = r_vir*constants.CM_PER_KPC
  m_vir_cgs = m_vir*constants.MSUN

  # Get v_c
  v_c_cgs = np.sqrt( constants.G_UNIV * m_vir_cgs / r_vir_cgs )

  # Convert to km/s
  v_c = v_c_cgs / constants.CM_PER_KM

  return v_c

########################################################################
# Metal Trends
########################################################################

def stellar_mass_metallicity_relation( m_star, apply_correction=True ):
    '''Mass-metallicity relation (MZR) in stars at z=0, taken from
        Gallazzi+2005 with a correction to the B-band weighted stellar
        metallicities by Peeples&Somerville2013
        (and put together in S3.1 of Peeples+14).

    Args:
        m_star (float or array-like) :
            Mass of the target galaxy at z=0 in units of Msun.

        apply_correction (bool) :
            If True, apply the correction from B-band weighted stellar
            metallicities to mass-weighted stellar metallicities calculated
            by Peeples&Somerville2013
    
    Returns:
        z (dict of np.ndarrays): Mass metallicity relation percentiles
    '''

    # Values from Table 2 of Gallazzi+2005
    m_star_gallazzi = 10.**np.arange( 8.91, 11.91 + .2, .2 )
    z_data = utilities.SmartDict({
        50 : 10.**np.array([
            -0.60,
            -0.61,
            -0.65,
            -0.61,
            -0.52,
            -0.41,
            -0.23,
            -0.11,
            -0.01,
            0.04,
            0.07,
            0.10,
            0.12,
            0.13,
            0.14,
            0.15,
        ]),
        16 : 10.**np.array([
            -1.11,
            -1.07,
            -1.10,
            -1.03,
            -0.97,
            -0.90,
            -0.80,
            -0.65,
            -0.41,
            -0.24,
            -0.14,
            -0.09,
            -0.06,
            -0.04,
            -0.03,
            -0.03,
        ]),
        84 : 10.**np.array([
            0.,
            0.,
            -0.05,
            -0.01,
            0.05,
            0.09,
            0.14,
            0.17,
            0.20,
            0.22,
            0.24,
            0.25,
            0.26,
            0.28,
            0.29,
            0.30,
        ]),
    })

    # Interpolate
    z = utilities.SmartDict( {} )
    for key, item in z_data.items():
        z[key] = interp.interp1d( m_star_gallazzi, item )( m_star )

    # Apply correction
    if apply_correction:
        for key, item in z.items():
            z[key] = 10.**( 1.08 * np.log10( item ) - 0.16 )

    return z

########################################################################

def ism_oxygen_mass_frac(
    m_star,
):
    '''Observations of Oxygen mass in the ISM compiled by Peeples+14.
    Data is from Kewley&Ellison2008.
    Args:
        m_star (float or array-like) :
            Mass of the target galaxy at z=0 in units of Msun.

    Returns:
        m_o (np.ndarray): log(O/H) + 12
    '''

    log_m_star = np.log10( m_star )
    log_moxy_mg = (
        27.76
        - 7.056 * log_m_star
        + 0.8184 * log_m_star**2.
        - 0.03029 * log_m_star**3.
    )

    return 10.**log_moxy_mg

########################################################################

def ism_mass_metallicity_relation(
    m_star,
    z_sun = 0.02,
):
    '''Get the ISM mass metallicity relation, as compiled by Peeples+14
    (S3.2.2). This is really just assuming that there is a fixed ratio
    between the oxygen mass and the stellar mass.

    Returns: ISM mass metallicity relation, in *mass fraction*
    '''

    return ism_oxygen_mass_frac( m_star ) / 0.44 * 1.366/16. * 1e-12 / z_sun

########################################################################

def galaxy_metal_mass(
    m_star,
    z_sun = 0.02,
    apply_stellar_z_correction = True,
    cold_gas_correction_factor = None,
    use_powerlaw_median_for_cold_gas = False,
    mass_sources = [ 'stars', 'ISM', 'dust' ],
):
    '''Mass of metals in galaxies, as assembled by Peeples+2014.

    Args:
        m_star (float or array-like) :
            Mass of the target galaxy at z=0 in units of Msun.

        z_sun (float) :
            Solar metallicity. Some (e.g. Peeples+14) use the Caffau+2011
            value of 0.0153, while others use the Asplund+2010 value of 0.014.
        apply_stellar_z_correction (bool) :
            If True, apply the correction from B-band weighted stellar
            metallicities to mass-weighted stellar metallicities calculated
            by Peeples&Somerville2013

        cold_gas_correction

    Returns:
        m_z (dict of dicts) :
            Means and percentages for each category, and the total.
    '''

    # Setup base dictionary
    m_z = {}

    for mass_source in mass_sources:

        # Stellar metal masses
        if mass_source == 'stars':

            # Get stellar metal mass out
            z_star = stellar_mass_metallicity_relation(
                m_star,
                apply_correction = apply_stellar_z_correction,
            )
            m_z[mass_source] = z_star * m_star * z_sun

        # ISM gas metal masses
        if mass_source == 'ISM':

            # Get gas mass out.
            m_gas = cold_gas_fraction(
                m_star,
                use_powerlaw_median = use_powerlaw_median_for_cold_gas,
            ) * m_star

            # Correct for only accounting for cold gas
            if cold_gas_correction_factor is not None:
                m_gas *= cold_gas_correction_factor

            # Get metallicity
            z_ism = ism_mass_metallicity_relation( m_star, z_sun )

            m_z[mass_source] = m_gas * z_ism

        # Dust mass
        if mass_source == 'dust':

            m_dust = ism_dust_mass( m_star )

            m_z[mass_source] = utilities.SmartDict( {
                16 : m_dust / 1.5,
                50 : m_dust,
                84 : m_dust * 1.5,
            } )

    # Total
    # Create the array
    m_z['total'] = utilities.SmartDict( {} )
    for key, item in m_z[mass_sources[0]].items():
        m_z['total'][key] = np.zeros( m_star.shape )
    # Sum up
    for mass_source in mass_sources:
        for key, item in m_z[mass_source].items():
            m_z['total'][key] += item

    return m_z

########################################################################

def galaxy_available_metals(
    m_star,
    alpha = 'Peeples14',
    y = 'Peeples14',
):
    '''Available metals in a galaxy, assuming a power law of the form
    y * m_star ** alpha . y is basically the yield, and alpha should usually
    be close to 1.

    Args:
        m_star (float or array-like) :
            Mass of the target galaxy at z=0 in units of Msun.

        alpha, y (float or str) :
            'Peeples14' uses a fit to digitized data from Peeples+14.

    Returns:
        galaxy_available_metals (same as m_star): Results
    '''

    if alpha == 'Peeples14':
        alpha = 1.014
    if y == 'Peeples14':
        y = 0.0479

    return y * m_star ** alpha

########################################################################

def cold_gas_fraction(
    m_star = None,
    use_powerlaw_median = True,
):
    '''Mass of cold gas in galaxies, as given by a compilation of data by
    Peeples+2014, consisting of data from McGaugh2005,2012, Leroy+2008,
    and Saintonge+2011.

    Args:
        m_star (float or array-like) :
            Mass of the target galaxy at z=0 in units of Msun.

    Returns:
        f_g (dict of np.ndarrays) :
            Cold gas fraction.
    '''

    m_star_peeples = 10.**np.array([
        6.7,
        7.1,
        7.6,
        8.2,
        8.6,
        9.1,
        9.6,
        10.1,
        10.6,
        11.0,
        11.4,
    ])
    f_g_data = {
        50 : np.array([
            7.4,
            3.7,
            5.6,
            5.8,
            3.6,
            1.3,
            0.6,
            0.5,
            0.23,
            0.17,
            0.067,
        ]),
        16 : np.array([
            4.6,
            2.1,
            2.2,
            4.8,
            2.6,
            1.0,
            0.32,
            0.083,
            0.056,
            0.082,
            0.057,
        ]),
        84 : np.array([
            12.0,
            19.5,
            15.7,
            8.4,
            4.3,
            2.2,
            1.7,
            1.2,
            0.52,
            0.34,
            0.12,
        ]),
    }

    # Interpolate
    if m_star is not None:
        f_g = utilities.SmartDict( {} )
        for key, item in f_g_data.items():
            f_g[key] = interp.interp1d( m_star_peeples, item )( m_star )
    else:
        f_g = utilities.SmartDict( f_g_data )

        return m_star_peeples, f_g

    # Replace the median with a fit powerlaw, if requested
    if use_powerlaw_median:
        f_g[50] = 10.**( -0.48 * np.log10( m_star ) + 4.39 )

    return f_g
        
########################################################################

def ism_dust_mass( m_star ):
    '''ISM dust mass, compiled by Peeples+2014, using data from Skibba+2011 and
    Draine+2007.

    Args:
        m_star (float or array-like) :
            Mass of the target galaxy at z=0 in units of Msun.

    Returns:
        ism_dust_mass (same as m_star):
            Units of Msun.
    '''

    return 10.**( 0.86 * np.log10( m_star ) - 1.31 )

########################################################################

def sneii_metal_budget( m_star, y_z_ii=0.030 ):
    '''Get the total mass of metal produced by SNe by z=0 in a galaxy with a
    z=0 stellar mass of m_star.
    This is taken from Peeple+2014, who use the star formation histories of
    Leitner+2012.

    Args:
        m_star (float or array-like) :
            Mass of the target galaxy at z=0 in units of Msun.

        y_z_ii (float) :
            Nucleosynthetic yield of all heavy 

    Returns:
        sneii_metal_budget (same as m_star):
            Units of Msun.
    '''

    return 1.2856 * y_z_ii * m_star ** 1.0146

########################################################################

def smhm_m_halo(log_m_star, z, paramfile=None ):
    '''Adapted from Behroozi+2018 (UniverseMachine EDR) by S.Wellons 7/11/18

    Args:
        log_m_star (float or array-like):
            log10 of the stellar mass in units of Msun.

        z (float) :
            Redshift.

        paramfile (str):
            What data to use.

    Returns:
        log_m_halo (same as log_m_star) :
            log10 of the halo mass in Msun
    '''

    if paramfile is None:
        dirname = os.path.dirname( __file__ )
        paramfile = os.path.join(
            dirname,
            "params",
            "smhm_true_med_params.txt"
        )

    log_m_halo = np.zeros(len(log_m_star))
    for i in range(0,len(log_m_star)):
        log_m_halo[i] = newton(_smhm_m_star, 12., args=(z, log_m_star[i]))
    return log_m_halo

def _smhm_m_star(log_m_halo, z, target):

    dirname = os.path.dirname( __file__ )
    paramfile = os.path.join(
        dirname,
        "params",
        "smhm_true_med_params.txt"
    )

    return smhm_m_star(np.asarray([log_m_halo]), z, paramfile=paramfile)[0] - target
            
########################################################################

def smhm_m_star(log_m_halo, z, paramfile=None):
    '''Adapted from Behroozi+2018 (UniverseMachine EDR) by S.Wellons 7/11/18

    Args:
        log_m_halo (same as log_m_star) :
            log10 of the halo mass in Msun

        z (float) :
            Redshift.

        paramfile (str):
            What data to use.

    Returns:
        log_m_star (float or array-like):
            log10 of the stellar mass in units of Msun.
    '''

    if paramfile is None:
        dirname = os.path.dirname( __file__ )
        paramfile = os.path.join(
            dirname,
            "params",
            "smhm_true_med_params.txt"
        )

    #Load params
    param_file = open(paramfile, "r") 
    param_list = []
    allparams = []
    for line in param_file:
        param_list.append(float((line.split(" "))[1]))
        allparams.append(line.split(" "))

    if (len(param_list) != 20):
        print("Parameter file not correct length.  (Expected 20 lines, got %d)." % len(param_list))
        quit()

    names = "EFF_0 EFF_0_A EFF_0_A2 EFF_0_Z M_1 M_1_A M_1_A2 M_1_Z ALPHA ALPHA_A ALPHA_A2 ALPHA_Z BETA BETA_A BETA_Z DELTA GAMMA GAMMA_A GAMMA_Z CHI2".split(" ");
    params = dict(zip(names, param_list))

    #Decide whether to print tex or evaluate SMHM parameter
    try:
        z = float(z)
    except:
        #print TeX
        for x in allparams[0:8:1]:
            sys.stdout.write('& $%.3f^{+%.3f}_{-%.3f}$' % tuple(float(y) for y in x[1:4]))
        sys.stdout.write("\\\\\n & & & ")
        for x in allparams[8:16:1]:
            sys.stdout.write('& $%.3f^{+%.3f}_{-%.3f}$' % tuple(float(y) for y in x[1:4]))
        sys.stdout.write("\\\\\n & & & ")    
        for x in allparams[16:19:1]:
            sys.stdout.write('& $%.3f^{+%.3f}_{-%.3f}$' % tuple(float(y) for y in x[1:4]))
        sys.stdout.write(' & %.0f' % float(allparams[19][1]))
        if (float(allparams[19][1])>200):
            sys.stdout.write('$\dag$')
        print('\\\\[2ex]')
        quit()

    #Print SMHM relation
    a = 1.0/(1.0+z)
    a1 = a - 1.0
    lna = math.log(a)
    zparams = {}
    zparams['m_1'] = params['M_1'] + a1*params['M_1_A'] - lna*params['M_1_A2'] + z*params['M_1_Z']
    zparams['sm_0'] = zparams['m_1'] + params['EFF_0'] + a1*params['EFF_0_A'] - lna*params['EFF_0_A2'] + z*params['EFF_0_Z']
    zparams['alpha'] = params['ALPHA'] + a1*params['ALPHA_A'] - lna*params['ALPHA_A2'] + z*params['ALPHA_Z']
    zparams['beta'] = params['BETA'] + a1*params['BETA_A'] + z*params['BETA_Z']
    zparams['delta'] = params['DELTA']
    zparams['gamma'] = 10**(params['GAMMA'] + a1*params['GAMMA_A'] + z*params['GAMMA_Z'])

    smhm_max = 14.5-0.35*z
    #print('#Log10(Mpeak/Msun) Log10(Median_SM/Msun) Log10(Median_SM/Mpeak)')
    #print('#Mpeak: peak historical halo mass, using Bryan & Norman virial overdensity.')
    #print('#Overall fit chi^2: %f' % params['CHI2'])
    # if (params['CHI2']>200):
    #     print('#Warning: chi^2 > 200 implies that not all features are well fit.  Comparison with the raw data (in data/smhm/median_raw/) is crucial.')
    log_m_star = log_m_halo*0.
    for i in range(len(log_m_halo)): #m in [x*0.05 for x in range(int(10.5*20),int(smhm_max*20+1),1)]:
        m = log_m_halo[i]
        dm = m-zparams['m_1'];
        dm2 = dm/zparams['delta'];
        sm = zparams['sm_0'] - math.log10(10**(-zparams['alpha']*dm) + 10**(-zparams['beta']*dm)) + zparams['gamma']*math.exp(-0.5*(dm2*dm2));
        log_m_star[i] = sm
        #print("%.2f %.6f %.6f" % (m,sm,sm-m))

    return log_m_star

########################################################################

def smhm_slope(log_m_halo, z, paramfile=None):
    '''Adapted from Behroozi+2018 (UniverseMachine EDR) by S.Wellons 7/11/18

    Args:
        log_m_halo (same as log_m_star) :
            log10 of the halo mass in Msun

        z (float) :
            Redshift.

        paramfile (str):
            What data to use.

    Returns:
        slope (float or array-like):
            Returns dlogMstar/dlogMhalo
    '''

    if paramfile is None:
        dirname = os.path.dirname( __file__ )
        paramfile = os.path.join(
            dirname,
            "params",
            "smhm_true_med_params.txt"
        )

    #Load params
    param_file = open(paramfile, "r") 
    param_list = []
    allparams = []
    for line in param_file:
        param_list.append(float((line.split(" "))[1]))
        allparams.append(line.split(" "))

    if (len(param_list) != 20):
        print("Parameter file not correct length.  (Expected 20 lines, got %d)." % len(param_list))
        quit()

    names = "EFF_0 EFF_0_A EFF_0_A2 EFF_0_Z M_1 M_1_A M_1_A2 M_1_Z ALPHA ALPHA_A ALPHA_A2 ALPHA_Z BETA BETA_A BETA_Z DELTA GAMMA GAMMA_A GAMMA_Z CHI2".split(" ");
    params = dict(zip(names, param_list))

    z = float(z)

    #Print SMHM relation
    a = 1.0/(1.0+z)
    a1 = a - 1.0
    lna = math.log(a)
    zparams = {}
    zparams['m_1'] = params['M_1'] + a1*params['M_1_A'] - lna*params['M_1_A2'] + z*params['M_1_Z']
    zparams['sm_0'] = zparams['m_1'] + params['EFF_0'] + a1*params['EFF_0_A'] - lna*params['EFF_0_A2'] + z*params['EFF_0_Z']
    zparams['alpha'] = params['ALPHA'] + a1*params['ALPHA_A'] - lna*params['ALPHA_A2'] + z*params['ALPHA_Z']
    zparams['beta'] = params['BETA'] + a1*params['BETA_A'] + z*params['BETA_Z']
    zparams['delta'] = params['DELTA']
    zparams['gamma'] = 10**(params['GAMMA'] + a1*params['GAMMA_A'] + z*params['GAMMA_Z'])

    smhm_max = 14.5-0.35*z
    slope = log_m_halo*0.
    for i in range(len(log_m_halo)): #m in [x*0.05 for x in range(int(10.5*20),int(smhm_max*20+1),1)]:
        m = log_m_halo[i]
        dm = m-zparams['m_1'];
        
        term1 = (zparams['alpha']*10.**(zparams['beta']*dm)+zparams['beta']*10.**(zparams['alpha']*dm))/(10.**(zparams['beta']*dm) + 10.**(zparams['alpha']*dm))
        term2 = -zparams['gamma']*dm*math.exp(-(dm/zparams['delta'])**2/2.)/zparams['delta']**2
        slope[i] = term1 + term2

    return slope



#!/usr/bin/env python
'''Subclass for analyzing particle data.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import galaxy_dive.analyze_data.generic_data as generic_data

########################################################################

class LOSData( generic_data.GenericData ):

  def __init__(self, data_p, **kw_args):
    '''Class for data that's a number of lines of sight in a grid shap.

    data_p : Parameters specifying the gridded snapshot file. Includes...
    "    "['sdir'] : Directory the snapshot is stored in
    "    "['snum'] : Snapshot number
    "    "['Nx'] : Number of grid cells on a side
    "    "['gridsize'] : How large the grid is
    "    "['ionized'] : How the ionization state is calculated.
                        'R13' : Ionization state calculated using Rahmati+13 fitting function
         "['comp_method'] : How the absorber component is calculated.
                            'all' : Entire LOS
                            'vshape' : Strongest absorber component in velocity space
                            'shape' : Strongest absorber component in physical space
    '''

    super(LOSData, self).__init__(data_p, **kw_args)

    self.retrieve_data()

    # State that we assume the data is centered at the start
    self.centered = True

    # State that we assume the data is centered at the start
    self.vel_centered = True

    # State that the data has already had the hubble flow accounted for
    self.hubble_corrected = True

    # Note units as necessary
    self.units['Z'] = 'solar'

  ########################################################################

  def retrieve_data(self):

    # Open file
    self.LOS_data_file_name = dataio.getLOSDataFilename(self.data_p['sdir'], self.data_p['Nx'], self.data_p['gridsize'], self.data_p['face'], \
                                                            self.data_p['comp_method'], self.data_p['ionized'], self.data_p['den_weight'])
    f = h5py.File(self.LOS_data_file_name, 'r')
    snapshot_data = f[str(self.data_p['snum'])]

    # Get the grid attributes
    self.data_attrs = {}
    for key in snapshot_data.attrs.keys():
      self.data_attrs[key] = snapshot_data.attrs[key]

    # Add the line of sight width (or cell length) to the data attributes
    self.data_attrs['cell_length'] = self.data_attrs['gridsize']/self.data_p['Nx']

    # Get the data
    self.data = {}
    for key in snapshot_data.keys():

      # Parse the keys and put in a more general format
      # All units should be the standard GIZMO output units
      # Gas Density
      if key == 'LOSDen':
        self.data['Den'] = snapshot_data[key][...]
      # Gas Neutral Hydrogen density
      elif key == 'LOSNHI':
        self.data['HIDen'] = snapshot_data[key][...]
      # Gas Neutral Hydrogen column density
      elif key == 'LOSHI':
        self.data['HI'] = snapshot_data[key][...]
      # Gas Hydrogen column density
      elif key == 'LOSH':
        self.data['H'] = snapshot_data[key][...]
      # Star Formation Rate
      elif key == 'LOSSFR':
        self.data['SFR'] = snapshot_data[key][...]
      # Temperature
      elif key == 'LOST':
        self.data['T'] = snapshot_data[key][...]
      # X Distance
      elif key == 'LOSRx':
        x = snapshot_data[key][...]
      # Y Distance
      elif key == 'LOSRy':
        y = snapshot_data[key][...]
      # Z Distance
      elif key == 'LOSRz':
        z = snapshot_data[key][...]
      # Radial Distance
      elif key == 'LOSR':
        self.data['R'] = snapshot_data[key][...]
      # Projected Distance
      elif key == 'LOSRr':
        self.data['Rho'] = snapshot_data[key][...]
      # X Velocity
      elif key == 'LOSVx':
        vx = snapshot_data[key][...]
      # Y Velocity
      elif key == 'LOSVy':
        vy = snapshot_data[key][...]
      # Z Velocity
      elif key == 'LOSVz':
        vz = snapshot_data[key][...]
      # Radial velocity
      elif key == 'LOSVr':
        self.data['Vr'] = snapshot_data[key][...]
      # Metallicity. Note that it's already in solar units in its main form, so we need to note that.
      elif key == 'LOSZ':
        self.data['Z'] = snapshot_data[key][...]
      # Standard deviation of the Metallicity (of the grid cells making up the LOS)
      elif key == 'LOSstdZ':
        self.data['StdZ'] = snapshot_data[key][...]
      # Log Metallicity (and averaged that way)
      elif key == 'LOSlogZ':
        self.data['LogZ'] = snapshot_data[key][...]
      # Standard deviation of Log Metallicity (of the grid cells making up the LOS)
      elif key == 'LOSstdlogZ':
        self.data['StdLogZ'] = snapshot_data[key][...]
      else:
        raise Exception('NULL key, key={}'.format(key))

    # Finish organizing velocity and position
    self.data['P'] = np.array([x, y, z])
    self.data['V'] = np.array([vx, vy, vz])

  ########################################################################

  def covering_fraction(self):
    '''Calculate the covering fraction for data satisfying all criteria.'''

    # Get the number of data fulfilling all criteria
    full_criteria_mask = self.get_total_mask()
    num_valid_data = float(np.invert(full_criteria_mask).sum())
  
    # Get the number of data fulfilling just the radius requirements
    num_impact_parameter_masks = 0
    for mask in self.masks:

      if mask['data_key'] == 'Rho':
        tot_mask = mask['mask']
        num_all_data = float(np.invert(tot_mask).sum())
        num_impact_parameter_masks += 1

      elif mask['data_key'] == 'Rhof':
        tot_mask = mask['mask']
        num_all_data = float(np.invert(tot_mask).sum())
        num_impact_parameter_masks += 1

    # Check that something didn't go wrong when getting the impact parameter masks
    if num_impact_parameter_masks != 1:
      raise Exception('num_impact_parameter_masks != 1, num_impact_parameter_masks = {}'.format(num_impact_parameter_masks))

    # Return the covering fraction
    self.f_cov = num_valid_data/num_all_data
    return self.f_cov

  ########################################################################

  def calc_mass(self):
    '''Calculate the mass for LOS cells.'''

    # Use the total H column density, and convert to total gas mass
    self.data['M'] = self.get_data('H')/0.75*constants.M_P
    self.data['M'] *= (self.data_attrs['cell_length']*constants.kpc_to_cm)**2.
    self.data['M'] /= constants.Msun_to_kg*1000. # Finish converting

########################################################################




#!/usr/bin/env python
'''Routines for reading in GADGET/GIZMO snapshots.

@author: Zach Hafen, Phil Hopkins, Claude-Andre Faucher-Giguere,
  and maybe others (this file's been passed around a lot).
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import glob
import h5py as h5py
import os

import numpy as np
import os.path
import scipy.interpolate as interpolate
import scipy.optimize as optimize
import math

########################################################################
########################################################################

def readsnap( sdir, snum, ptype, load_additional_ids=0, snapshot_name='snapshot', extension='.hdf5', h0=0,cosmological=0, skip_bh=0, four_char=0, header_only=0, loud=0):
    '''Reads in a particle snapshot.

    Args:
      sdir (str): Simulation directory to load from.

      snum (int): Snapshot to load.

      ptype (int): Particle type to load. Typically, 0=gas, 1=hi-res dark matter, 2=low-res dark matter, 4=stars.

      load_additional_ids (bool): Whether or not to also load the fields 'ParticleChildIDsNumber' and 'ParticleIDGenerationNumber'.

      snapshot_name (str): The base name for the snapshot.

      extension (str): The extension for the snapshot.

      h0 (bool): If True, use present day values.

      cosmological (bool): If True, convert from comoving to physical units, and return in physical units, with factors of h accounted for.

      skip_bh (bool): If True, don't read BH-specific fields for particles of type 5.

      four_char (bool): If True, then snapshot files have 4 digits, instead of three (e.g. 0020 instead of 020)

      header_only (bool): If True, return only the header, in stead of the full data.

      loud (bool): If True, print additional information.

    Returns:
      P (dict): A dictionary of numpy arrays containing the particle data.
    '''
    
    if (ptype<0): return {'k':-1};
    if (ptype>5): return {'k':-1};

    #print("just past ptype check")

    fname,fname_base,fname_ext = check_if_filename_exists(sdir,snum,\
        snapshot_name=snapshot_name,extension=extension,four_char=four_char)
    if(fname=='NULL'): 
      raise Exception( "Couldn't read snapshot {} in {}.\nCheck if this is the right location.".format( snum, sdir ) )
    if(loud==1): print('loading file : '+fname)

    ## open file and parse its header information
    nL = 0 # initial particle point to start at 
    if(fname_ext=='.hdf5'):
        file = h5py.File(fname,'r') # Open hdf5 snapshot file
        header_master = file["Header"] # Load header dictionary (to parse below)
        header_toparse = header_master.attrs
    else:
        file = open(fname) # Open binary snapshot file
        header_toparse = load_gadget_binary_header(file)

    npart = header_toparse["NumPart_ThisFile"]
    massarr = header_toparse["MassTable"]
    time = header_toparse["Time"]
    redshift = header_toparse["Redshift"]
    flag_sfr = header_toparse["Flag_Sfr"]
    flag_feedbacktp = header_toparse["Flag_Feedback"]
    npartTotal = header_toparse["NumPart_Total"]
    flag_cooling = header_toparse["Flag_Cooling"]
    numfiles = header_toparse["NumFilesPerSnapshot"]
    boxsize = header_toparse["BoxSize"]
    omega_matter = header_toparse["Omega0"]
    omega_lambda = header_toparse["OmegaLambda"]
    hubble = header_toparse["HubbleParam"]
    flag_stellarage = header_toparse["Flag_StellarAge"]
    flag_metals = header_toparse["Flag_Metals"]

    hinv=1.
    if (h0==1):
        hinv=1./hubble
    ascale=1.
    if (cosmological==1):
        ascale=time
        hinv=1./hubble
    if (cosmological==0): 
        time*=hinv

    # CAFG: changed order of the following two lines, so that header
    # info is returned even if no particles.
    if (header_only==1):
      file.close()
      return {'k':0,'time':time,'hubble':hubble,'redshift':redshift, 'npart':npartTotal[ptype]}
    if (npartTotal[ptype]<=0): file.close(); return {'k':-1};


    # initialize variables to be read
    pos=np.zeros([npartTotal[ptype],3],dtype=float)
    vel=np.copy(pos)
    ids=np.zeros([npartTotal[ptype]],dtype=np.long)
    if load_additional_ids:
      child_ids=np.zeros([npartTotal[ptype]],dtype=np.long)
      id_gens=np.zeros([npartTotal[ptype]],dtype=np.long)
    mass=np.zeros([npartTotal[ptype]],dtype=float)
    potential = np.copy( mass )
    if (ptype==0):
        ugas=np.copy(mass)
        rho=np.copy(mass)
        hsml=np.copy(mass) 
        if (flag_cooling>0): 
            nume=np.copy(mass)
            numh=np.copy(mass)
        if (flag_sfr>0): 
            sfr=np.copy(mass)
    if (ptype==0 or ptype==4) and (flag_metals > 0):
        metal=np.zeros([npartTotal[ptype],flag_metals],dtype=float)
    if (ptype==4) and (flag_sfr>0) and (flag_stellarage>0):
        stellage=np.copy(mass)
    if (ptype==5) and (skip_bh==0):
        bhmass=np.copy(mass)
        bhmdot=np.copy(mass)

    # loop over the snapshot parts to get the different data pieces
    for i_file in range(numfiles):
        if (numfiles>1):
            file.close()
            fname = fname_base+'.'+str(i_file)+fname_ext
            if(fname_ext=='.hdf5'):
                file = h5py.File(fname,'r') # Open hdf5 snapshot file
            else:
                file = open(fname) # Open binary snapshot file
                header_toparse = load_gadget_binary_header(file)
                
        if (fname_ext=='.hdf5'):
            input_struct = file
            npart = file["Header"].attrs["NumPart_ThisFile"]
            ptype_full_name = "PartType{}".format( ptype )
            bname = ptype_full_name + "/"
        else:
            npart = header_toparse['NumPart_ThisFile']
            input_struct = load_gadget_binary_particledat(file, header_toparse, ptype, skip_bh=skip_bh)
            bname = ''
            
        
        # now do the actual reading
        
        # CAFG: there can be an error here when particles of a certain
        # type (e.g., stars) are only in a subset of all files that
        # constitute the snapshot. Added 'if' on npart[ptype]>0.

        if npart[ptype]>0:
            nR=nL + npart[ptype]
            pos[nL:nR,:]=input_struct[bname+"Coordinates"]
            vel[nL:nR,:]=input_struct[bname+"Velocities"]
            ids[nL:nR]=input_struct[bname+"ParticleIDs"]
            mass[nL:nR]=massarr[ptype]

            potential_stored = 'Potential' in input_struct[ptype_full_name].keys()
            if potential_stored:
                potential[nL:nR] = input_struct[bname+"Potential"]

            if (massarr[ptype] <= 0.):
                mass[nL:nR]=input_struct[bname+"Masses"]
            if (ptype==0):
                ugas[nL:nR]=input_struct[bname+"InternalEnergy"]
                rho[nL:nR]=input_struct[bname+"Density"]
                hsml[nL:nR]=input_struct[bname+"SmoothingLength"]
                if (flag_cooling > 0): 
                    nume[nL:nR]=input_struct[bname+"ElectronAbundance"]
                    numh[nL:nR]=input_struct[bname+"NeutralHydrogenAbundance"]
                if (flag_sfr > 0):
                    sfr[nL:nR]=input_struct[bname+"StarFormationRate"]
            if (ptype==0 or ptype==4) and (flag_metals > 0):
                metal_t=input_struct[bname+"Metallicity"]
                if (flag_metals > 1):
                    if (metal_t.shape[0] != npart[ptype]): 
                        metal_t=np.transpose(metal_t)
                else:
                    metal_t=np.reshape(np.array(metal_t),(np.array(metal_t).size,1))
                metal[nL:nR,:]=metal_t
            if (ptype==4) and (flag_sfr>0) and (flag_stellarage>0):
                stellage[nL:nR]=input_struct[bname+"StellarFormationTime"]
            if (ptype==5) and (skip_bh==0):
                bhmass[nL:nR]=input_struct[bname+"BH_Mass"]
                bhmdot[nL:nR]=input_struct[bname+"BH_Mdot"]

            if load_additional_ids:
              child_ids[nL:nR]=input_struct[bname+"ParticleChildIDsNumber"]
              id_gens[nL:nR]=input_struct[bname+"ParticleIDGenerationNumber"]

            nL = nR # sets it for the next iteration	

        ## correct to same ID as original gas particle for new stars, if bit-flip applied
    if ((np.min(ids)<0) | (np.max(ids)>1.e9)):
        bad = (ids < 0) | (ids > 1.e9)
        ids[bad] += (int(1) << 31)

    # do the cosmological conversions on final vectors as needed
    pos *= hinv*ascale # snapshot units are comoving
    boxsize *= hinv*ascale
    mass *= hinv
    vel *= np.sqrt(ascale) # remember gadget's weird velocity units!
    if (ptype==0):
        rho *= (hinv/((ascale*hinv)**3))
        hsml *= hinv*ascale
    if (ptype==4) and (flag_sfr>0) and (flag_stellarage>0) and (cosmological==0):
        stellage *= hinv
    if (ptype==5) and (skip_bh==0):
        bhmass *= hinv

    file.close();
    if (ptype==0):
       ret_dict = {'hubble':hubble,'boxsize':boxsize,'time':time,'redshift':redshift,'flag_sfr':flag_sfr,'flag_feedbacktp':flag_feedbacktp,'flag_cooling':flag_cooling,'omega_matter':omega_matter,'omega_lambda':omega_lambda,'flag_stellarage':flag_stellarage,'flag_metals':flag_metals,'k':1,'p':pos,'v':vel,'m':mass,'id':ids,'u':ugas,'rho':rho,'h':hsml,'ne':nume,'nh':numh,'sfr':sfr}
       if flag_metals>0:
         ret_dict['z'] = metal
    elif (ptype==4):
        ret_dict = {'hubble':hubble,'boxsize':boxsize,'time':time,'redshift':redshift,'flag_sfr':flag_sfr,'flag_feedbacktp':flag_feedbacktp,'flag_cooling':flag_cooling,'omega_matter':omega_matter,'omega_lambda':omega_lambda,'flag_stellarage':flag_stellarage,'flag_metals':flag_metals,'k':1,'p':pos,'v':vel,'m':mass,'id':ids,'age':stellage}
        if flag_metals>0:
          ret_dict['z'] = metal
    elif (ptype==5) and (skip_bh==0):
        ret_dict = {'hubble':hubble,'boxsize':boxsize,'time':time,'redshift':redshift,'flag_sfr':flag_sfr,'flag_feedbacktp':flag_feedbacktp,'flag_cooling':flag_cooling,'omega_matter':omega_matter,'omega_lambda':omega_lambda,'flag_stellarage':flag_stellarage,'flag_metals':flag_metals,'k':1,'p':pos,'v':vel,'m':mass,'id':ids,'mbh':bhmass,'mdot':bhmdot}

    else:
      ret_dict = {'hubble':hubble,'boxsize':boxsize,'time':time,'redshift':redshift,'flag_sfr':flag_sfr,'flag_feedbacktp':flag_feedbacktp,'flag_cooling':flag_cooling,'omega_matter':omega_matter,'omega_lambda':omega_lambda,'flag_stellarage':flag_stellarage,'flag_metals':flag_metals,'k':1,'p':pos,'v':vel,'m':mass,'id':ids,}

    # Include optional ids
    if load_additional_ids:
      ret_dict['child_id'] = child_ids
      ret_dict['id_gen'] = id_gens

    if potential_stored:
      ret_dict['potential'] = potential/ascale # Convert to physical km^2/s^2

    return ret_dict

########################################################################

def check_if_filename_exists(sdir,snum,snapshot_name='snapshot',extension='.hdf5',four_char=0):
    '''Check if a file name exists, and if so return information about it.

    Args:
      sdir (str): Simulation directory to load from.
      snum (int): Snapshot to load.
      snapshot_name (str): The base name for the snapshot.
      extension (str): The extension for the snapshot.
      four_char (bool): If True, then snapshot files have 4 digits, instead of three (e.g. 0020 instead of 020)

    Returns:
      fname_found (bool): If true, the filename was found.
      fname_base_found (str): The base for the found filename.
      fname_ext (str): The extension for the found filename.
    '''
    for extension_touse in [extension,'.bin','']:
        fname=sdir+'/'+snapshot_name+'_'
        ext='00'+str(snum);
        if (snum>=10): ext='0'+str(snum)
        if (snum>=100): ext=str(snum)
        if (four_char==1): ext='0'+ext
        if (snum>=1000): ext=str(snum)
        fname+=ext
        fname_base=fname

        s0=sdir.split("/"); snapdir_specific=s0[len(s0)-1];
        if(len(snapdir_specific)<=1): snapdir_specific=s0[len(s0)-2];

        ## try several common notations for the directory/filename structure
        fname=fname_base+extension_touse;
        if not os.path.exists(fname): 
            ## is it a multi-part file?
            fname=fname_base+'.0'+extension_touse;
        if not os.path.exists(fname): 
            ## is the filename 'snap' instead of 'snapshot'?
            fname_base=sdir+'/snap_'+ext; 
            fname=fname_base+extension_touse;
        if not os.path.exists(fname): 
            ## is the filename 'snap' instead of 'snapshot', AND its a multi-part file?
            fname=fname_base+'.0'+extension_touse;
        if not os.path.exists(fname): 
            ## is the filename 'snap(snapdir)' instead of 'snapshot'?
            fname_base=sdir+'/snap_'+snapdir_specific+'_'+ext; 
            fname=fname_base+extension_touse;
        if not os.path.exists(fname): 
            ## is the filename 'snap' instead of 'snapshot', AND its a multi-part file?
            fname=fname_base+'.0'+extension_touse;
        if not os.path.exists(fname): 
            ## is it in a snapshot sub-directory? (we assume this means multi-part files)
            fname_base=sdir+'/snapdir_'+ext+'/'+snapshot_name+'_'+ext; 
            fname=fname_base+'.0'+extension_touse;
        if not os.path.exists(fname): 
            ## is it in a snapshot sub-directory AND named 'snap' instead of 'snapshot'?
            fname_base=sdir+'/snapdir_'+ext+'/'+'snap_'+ext; 
            fname=fname_base+'.0'+extension_touse;
        if not os.path.exists(fname): 
            ## wow, still couldn't find it... ok, i'm going to give up!
            fname_found = 'NULL'
            fname_base_found = 'NULL'
            fname_ext = 'NULL'
            continue;
        fname_found = fname;
        fname_base_found = fname_base;
        fname_ext = extension_touse
        break; # filename does exist! 
    return fname_found, fname_base_found, fname_ext;



def load_gadget_binary_header(f):
    '''Get the header information for a binary snapshot file.
  
    Args:
      f (binary file): The file to read the information from.

    Returns:
      header_info (dict): A dictionary of info contained in the header.
    '''
    ### Read header.
    import array
    # Skip 4-byte integer at beginning of header block.
    f.read(4)
    # Number of particles of each type. 6*unsigned integer.
    Npart = array.array('I')
    Npart.fromfile(f, 6)
    # Mass of each particle type. If set to 0 for a type which is present, 
    # individual particle masses from the 'mass' block are used instead.
    # 6*double.
    Massarr = array.array('d')
    Massarr.fromfile(f, 6)
    # Expansion factor (or time, if non-cosmological sims) of output. 1*double. 
    a = array.array('d')
    a.fromfile(f, 1)
    a = a[0]
    # Redshift of output. Should satisfy z=1/a-1. 1*double.
    z = array.array('d')
    z.fromfile(f, 1)
    z = float(z[0])
    # Flag for star formation. 1*int.
    FlagSfr = array.array('i')
    FlagSfr.fromfile(f, 1)
    # Flag for feedback. 1*int.
    FlagFeedback = array.array('i')
    FlagFeedback.fromfile(f, 1)
    # Total number of particles of each type in the simulation. 6*int.
    Nall = array.array('i')
    Nall.fromfile(f, 6)
    # Flag for cooling. 1*int.
    FlagCooling = array.array('i')
    FlagCooling.fromfile(f, 1)
    # Number of files in each snapshot. 1*int.
    NumFiles = array.array('i')
    NumFiles.fromfile(f, 1)
    # Box size (comoving kpc/h). 1*double.
    BoxSize = array.array('d')
    BoxSize.fromfile(f, 1)
    # Matter density at z=0 in units of the critical density. 1*double.
    Omega0 = array.array('d')
    Omega0.fromfile(f, 1)
    # Vacuum energy density at z=0 in units of the critical density. 1*double.
    OmegaLambda = array.array('d')
    OmegaLambda.fromfile(f, 1)
    # Hubble parameter h in units of 100 km s^-1 Mpc^-1. 1*double.
    h = array.array('d')
    h.fromfile(f, 1)
    h = float(h[0])
    # Creation times of stars. 1*int.
    FlagAge = array.array('i')
    FlagAge.fromfile(f, 1)
    # Flag for metallicity values. 1*int.
    FlagMetals = array.array('i')
    FlagMetals.fromfile(f, 1)

    # For simulations that use more than 2^32 particles, most significant word 
    # of 64-bit total particle numbers. Otherwise 0. 6*int.
    NallHW = array.array('i')
    NallHW.fromfile(f, 6)

    # Flag that initial conditions contain entropy instead of thermal energy
    # in the u block. 1*int.
    flag_entr_ics = array.array('i')
    flag_entr_ics.fromfile(f, 1)

    # Unused header space. Skip to particle positions.
    f.seek(4+256+4+4)

    return {'NumPart_ThisFile':Npart, 'MassTable':Massarr, 'Time':a, 'Redshift':z, \
    'Flag_Sfr':FlagSfr[0], 'Flag_Feedback':FlagFeedback[0], 'NumPart_Total':Nall, \
    'Flag_Cooling':FlagCooling[0], 'NumFilesPerSnapshot':NumFiles[0], 'BoxSize':BoxSize[0], \
    'Omega0':Omega0[0], 'OmegaLambda':OmegaLambda[0], 'HubbleParam':h, \
    'Flag_StellarAge':FlagAge[0], 'Flag_Metals':FlagMetals[0], 'Nall_HW':NallHW, \
    'Flag_EntrICs':flag_entr_ics[0]}

########################################################################
########################################################################

def gas_mu(num_e):
    '''Calculates the mean weight from gadget.'''

    XH=0.76 # we track this with metal species now, could do better...
    yhelium=(1.-XH)/(4.*XH)
    return (1.+4.*yhelium)/(1.+yhelium+num_e)

########################################################################

def gas_temperature(u, num_e, keV=0):
    '''Returns gas particles temperature in Kelvin. Originally from PFH's gadget lib (maybe somewhere else before?).'''

    g_gamma= 5.0/3.0
    g_minus_1= g_gamma-1.0
    PROTONMASS = 1.6726e-24
    BoltzMann_ergs= 1.3806e-16
    UnitMass_in_g= 1.989e43 # 1.0e10 solar masses
    UnitEnergy_in_cgs= 1.989e53
    # note gadget units of energy/mass = 1e10 ergs/g,
    # this comes into the formula below

    mu = gas_mu(num_e);
    MeanWeight= mu*PROTONMASS
    Temp= MeanWeight/BoltzMann_ergs * g_minus_1 * u * 1.e10

    # do we want units of keV?  (0.001 factor converts from eV to keV)
    if (keV==1):
        BoltzMann_keV = 8.617e-8;
        Temp *= BoltzMann_keV;

    return Temp

########################################################################
########################################################################
'''Unfinished below this line. TODO.'''

def read_snapshot( sdir, snum, ptype, load_additional_ids=False, cosmological=True, ):
  '''
  Note: This is created as a function, not a class, because I'm encountering memory issues that may be related to
  classes persisting in memory. -- Zach Hafen
  '''

  filepaths = get_snapshot_filepaths( sdir, snum )

  if cosmological:
    convert_data_to_physical_units()

  return data, data_attrs

########################################################################

def get_snapshot_filepaths( sdir, snum ):
  '''Get the filepath(s) to a snapshot located in a specified directory.

  Args:
    sdir (str) : Directory containing the simulation snapshot.
    snum (int) : Snapshot number.

  Returns:
    filepaths (list of strs) : Filepath(s) to the snapshot files.
  '''

  filecase_filepath = os.path.join( sdir, 'snapshot_{:03d}.hdf5'.format( snum ) )
  is_file = os.path.isfile( filecase_filepath )

  dir_filepath = os.path.join( sdir, 'snapdir_{:03d}'.format( snum ) )
  is_dir = os.path.isdir( dir_filepath )

  if not ( is_file or is_dir):
    raise NameError( "Snapshot {:03d} not found in {}".format( snum, sdir ) )
    
  if is_file:
    return [ filecase_filepath, ]
  elif is_dir:
    return glob.glob( '{}/*'.format( dir_filepath ) )

########################################################################

def read_snapshot_files( filepaths ):


  data_attrs = {}
  for i, filepath in enumerate( filepaths ):
    with h5py.File( filepath, 'r' ) as f:

      # First time data
      if i == 0:

        pass

      # Get the attrs/header information

########################################################################

def setup_data_struct( ptype, flag_metals ):
  '''
  Args:
    ptype (int) : Particle type to load.
    flag_metals (int) : How many metal species there are.
  '''

  # Setup data storage.
  data = {
    'P' : np.zeros( [ n_particles, 3], dtype=np.float64 ),
    'V' : np.zeros( [ n_particles, 3], dtype=np.float32 ),
    'ID' : np.zeros( n_particles, dtype=np.long ),
    'M' : np.zeros( n_particles, dtype=np.float32 ),
  }
  if load_additional_ids:
    data['ChildID'] = np.zeros( n_particles, dtype=np.long )
    data['IDGen'] = np.zeros( n_particles, dtype=np.long )

  # Information that varies depending on ptype
  if ptype == 0:
    data['U'] = np.zeros( n_particles, dtype=np.float32 )
    data['Den'] = np.zeros( n_particles, dtype=np.float32 )
    data['XHI'] = np.zeros( n_particles, dtype=np.float32 )
  if ( ptype == 0 ) or ( ptype == 4 ):
    data['Z'] = np.zeros( (n_particles, flag_metals), dtype=np.float32 )
  if ( ptype == 4):
    data['BirthTime'] = np.zeros( n_particles, dtype=np.float32 )

  return data



#!/usr/bin/env python
'''Class for analyzing simulation data.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

# Base python imports
import copy
import numpy as np
import numpy.testing as npt
import pandas as pd
import scipy
import scipy.signal as signal

# Imports from my own stuff
import galaxy_dive.analyze_data.ahf as analyze_ahf
import galaxy_dive.read_data.ahf as read_ahf
import galaxy_dive.utils.astro as astro
import galaxy_dive.utils.constants as constants
import galaxy_dive.utils.data_operations as data_operations
import galaxy_dive.analyze_data.generic_data as generic_data
import galaxy_dive.utils.utilities as utilities

########################################################################
########################################################################


class SimulationData( generic_data.GenericData ):
    '''Class for handling simulation data.'''

    @utilities.store_parameters
    def __init__(
        self,
        data_dir = None,
        halo_data_dir = None,
        ahf_index = None,
        averaging_frac = 5.,
        length_scale_used = 'Rstar0.5',
        z_sun = constants.Z_MASSFRAC_SUN,
        halo_data_retrieved = False,
        centered = False,
        vel_centered = False,
        hubble_corrected = False,
        ahf_tag = 'smooth',
        main_halo_id = 0,
        center_method = 'halo',
        vel_center_method = 'halo',
        store_ahf_reader = False,
        **kwargs
    ):
        '''Initialize.

        Args:
            data_dir (str) :
                 Directory the simulation is contained in.

            halo_data_dir (str) :
                 Directory simulation analysis is contained in. Defaults to data_dir

            ahf_index (str) :
                 What to index the snapshots by.
                 Should be the last snapshot in the simulation *if*
                 AHF was run backwards from the last snapshot.
                 Required to put in manually to avoid easy mistakes.

            averaging_frac (dict) :
                There are some averaged quantities (e.g. COM velocity, total
                angular momentum) that we consider. This is the fraction of
                length_scale_used within which we calculate these.

            length_scale_used (str) :
                 What length scale to use for the simulation. Will be used to put lengths in fractions.

            z_sun (float) :
                Used mass fraction for solar metallicity.
                Options...
                'r_scale' : Scale radius.
                'R_vir' : Virial radius.

            halo_data_retrieved (bool) :
                Whether or not we retrieved relevant values from the AHF halo data.

            centered (bool):
                Whether or not the coordinates are centered on the galaxy of choice at the start.

            vel_centered (bool) :
                Whether or not the velocities are relative to the galaxy of choice at the start.

            hubble_corrected (bool) :
                Whether or not the velocities have had the Hubble flow added
                (velocities must be centered).

            ahf_tag (str) :
                Identifying tag for the ahf merger tree halo files,
                looks for ahf files of type 'halo_00000_{}.dat'.format( tag ).

            main_halo_id (int) :
                 What is the halo ID of the main galaxy in the simulation?

            center_method (str or np.array) :
                How to center the coordinates. Options...
                'halo' (default) :
                     Centers the dataset on the main halo (main_halo_id) using AHF halo data.
                np.array :
                     Array of coordinates on which to center the data

            vel_center_method (str or np.array of size 3) :
                How to center the velocity coordinates, i.e. what the
                velocity is relative to. Options are...
                'halo' (default) :
                    Sets velocity relative to the main halo (main_halo_id)
                    using AHF halo data.
                np.array of size 3 :
                    Centers the dataset on this coordinate.

        Keyword Args:
            function_args (dict): Dictionary of args used to specify an arbitrary function with which to generate data.
        '''

        # Make sure that all the arguments have been specified.
        for attr in vars( self ).keys():
            if attr == 'kwargs':
                continue
            if getattr( self, attr ) is None:

                # Set the analysis dir to data_dir if not given
                if attr == 'halo_data_dir':
                    self.halo_data_dir = self.data_dir

                elif attr == 'ahf_index':
                    continue

                else:
                    raise Exception( '{} not specified'.format( attr ) )

        # By definition, the halo data should not be retrieved when the class is first initiated.
        self.halo_data_retrieved = False

        super( SimulationData, self ).__init__( z_sun=z_sun, **kwargs )

    ########################################################################
    # Properties
    ########################################################################

    @property
    def length_scale( self ):
        '''Property for fiducial simulation length scale.'''

        if self.length_scale_used == 'R_vir':
            return self.r_vir
        elif self.length_scale_used == 'r_scale':
            return self.r_scale
        else:
            return self.halo_data.get_mt_data(
                self.length_scale_used,
                snums = self.snum,
                return_values_only = False, # This is because we're only getting one value out
                a_power = 1.,
            ) / self.data_attrs['hubble']

    ########################################################################

    @property
    def velocity_scale( self ):
        '''Property for fiducial simulation velocity scale.'''

        return self.v_c

    ########################################################################

    @property
    def base_data_shape( self ):
        '''Property for simulation redshift.'''

        if not hasattr( self, '_base_data_shape' ):

            # Use Density as the default data we assume will usually be there.
            if 'Den' in self.data:
                self._base_data_shape = self.data['Den'].shape
            # If it doesn't have density, it might have mass
            elif 'M' in self.data:
                self._base_data_shape = self.data['M'].shape
            else:
                raise Exception( "No data key to base shape off of." )

        return self._base_data_shape

    @base_data_shape.setter
    def base_data_shape( self, value ):
        '''Setting function for simulation redshift property.'''

        # If we try to set it, make sure that if it already exists we don't change it.
        if hasattr( self, '_base_data_shape' ):
            assert self._base_data_shape == value

        else:
            self._base_data_shape = value

    ########################################################################

    @property
    def halo_data( self ):
        '''Halo data used.
        '''

        if not hasattr( self, '_halo_data' ):

            self._halo_data = analyze_ahf.HaloData(
                self.halo_data_dir,
                tag = self.ahf_tag,
                index = self.ahf_index
            )

        return self._halo_data

    ########################################################################

    @property
    def redshift( self ):
        '''Property for simulation redshift.'''

        if not hasattr( self, '_redshift' ):

            # Try to get it from the attributes.
            if 'redshift' in self.data_attrs:
                self._redshift = self.data_attrs['redshift']
            elif hasattr( self, 'data' ):
                if 'redshift' in self.data:
                    self._redshift = self.data['redshift']
            # If not, retrieve halo data, which should set it.
            # In fact, if we call self.retrieve_halo_data() somewhere else and we already set redshift by getting it from
            # the attributes, it will check that it matches.
            else:
                self.retrieve_halo_data()

        return self._redshift

    @redshift.setter
    def redshift( self, value ):
        '''Setting function for simulation redshift property.'''

        # If we try to set it, make sure that if it already exists we don't change it.
        if hasattr( self, '_redshift' ):

            if isinstance( value, np.ndarray ) or isinstance( self._redshift, np.ndarray ):

                is_nan = np.any( [ np.isnan( value ), np.isnan( self._redshift ) ], axis=1 )
                not_nan_inds = np.where( np.invert( is_nan ) )[0]

                test_value = np.array(value)[not_nan_inds]  # Cast as np.ndarray because Pandas arrays can cause trouble.
                test_existing_value = np.array(self._redshift)[not_nan_inds]
                npt.assert_allclose( test_value, test_existing_value, atol=1e-5 )

                self._redshift = value

            else:
                npt.assert_allclose( value, self._redshift, atol=1e-5 )

        else:
            self._redshift = value

    ########################################################################

    @property
    def r_vir( self ):
        '''Property for virial radius.'''

        if not hasattr( self, '_r_vir' ):
            self.retrieve_halo_data()

        return self._r_vir

    @r_vir.setter
    def r_vir( self, value ):

        # If we try to set it, make sure that if it already exists we don't change it.
        if hasattr( self, '_r_vir' ):
            npt.assert_allclose( value, self._r_vir )

        else:
            self._r_vir = value

    ########################################################################

    @property
    def r_scale( self ):
        '''Property for scale radius.'''

        if not hasattr( self, '_r_scale' ):
            self.retrieve_halo_data()

        return self._r_scale

    @r_scale.setter
    def r_scale( self, value ):

        # If we try to set it, make sure that if it already exists we don't change it.
        if hasattr( self, '_r_scale' ):
            npt.assert_allclose( value, self._r_scale )

        else:
            self._r_scale = value

    ########################################################################

    @property
    def v_c( self ):
        '''Property for circular velocity.'''

        if not hasattr( self, '_v_c' ):
            self.retrieve_halo_data()

        return self._v_c

    @v_c.setter
    def v_c( self, value ):

        # If we try to set it, make sure that if it already exists we don't change it.
        if hasattr( self, '_v_c' ):
            npt.assert_allclose( value, self._v_c )

        else:
            self._v_c = value

    ########################################################################

    @property
    def hubble_z( self ):
        '''Property for the hubble function at specified redshift.'''

        if not hasattr( self, '_hubble_z' ):
            self._hubble_z = astro.hubble_parameter(
                self.redshift,
                h=self.data_attrs['hubble'],
                omega_matter=self.data_attrs['omega_matter'],
                omega_lambda=self.data_attrs['omega_lambda'],
                units='km/s/kpc'
            )

        return self._hubble_z

    ########################################################################
    # Overall changes to the data
    ########################################################################

    def center_coords( self ):
        '''Change the location of the origin, if the data isn't already centered.

        Modifies:
            self.data['P'] : Shifts the coordinates to the center.
        '''

        if self.centered:
            return

        if isinstance( self.center_method, np.ndarray ):
            self.origin = copy.copy( self.center_method )

        elif self.center_method == 'halo':
            self.retrieve_halo_data()
            self.origin = copy.copy( self.halo_coords )

        else:
            raise KeyError( "Unrecognized center_method, {}".format( self.center_method ) )

        # Do it like this because we don't know the shape of self.data['P'][0]
        for i in range( 3 ):
            self.data['P'][i] -= self.origin[i]

        # Note that we're now centered
        self.centered = True

    ########################################################################

    def center_vel_coords( self ):
        '''Get velocity coordinates to center on the main halo.

        Modifies:
            self.data['V'] : Makes all velocities relative to self.vel_origin
        '''

        if self.vel_centered:
            return

        if isinstance( self.vel_center_method, np.ndarray ):
            self.vel_origin = copy.copy( self.vel_center_method )

        elif self.vel_center_method == 'halo':
            self.retrieve_halo_data()
            self.vel_origin = copy.copy( self.halo_velocity )

        else:
            raise KeyError( "Unrecognized vel_center_method, {}".format( self.vel_center_method ) )

        # Do it like this because we don't know the shape of self.data['V'][0]
        for i in range( 3 ):
            self.data['V'][i] -= self.vel_origin[i]

        self.vel_centered = True

    ########################################################################

    def add_hubble_flow( self ):
        '''Correct for hubble flow movement.

        Modifies:
            self.data['V'] : Accounts for hubble flow, relative to origin
        '''

        if self.hubble_corrected:
            return

        self.center_vel_coords()

        # Handle weird formatting that happens when using data frames.
        if isinstance( self.hubble_z, pd.Series ):
            self.data['V'] += self.get_data( 'P' ) * self.hubble_z.values
        else:
            self.data['V'] += self.get_data( 'P' ) * self.hubble_z

        self.hubble_corrected = True

    ########################################################################
    # Get Data
    ########################################################################

    def get_data( self, data_key, sl=None ):
        '''Get the data from within the class. Only for getting the data. No post-processing or changing the data
        (putting it in particular units, etc.) The idea is to calculate necessary quantities as the need arises,
        hence a whole function for getting the data.

        Args:
            data_key (str) : Key in the data dictionary for the key we want to get
            sl (slice) : Slice of the data, if requested.

        Returns:
            data (np.ndarray) : Requested data.
        '''

        # Loop through, handling issues
        n_tries = 10
        for i in range( n_tries ):
            try:

                # Positions
                if self.key_parser.is_position_key( data_key ):
                    data = self.get_position_data( data_key )

                # Velocities
                elif self.key_parser.is_velocity_key( data_key ):

                    data = self.get_velocity_data( data_key )

                # Arbitrary functions of the data
                elif data_key == 'Function':

                    raise Exception( "TODO: Test this" )

                    # Use the keys to get the data
                    function_data_keys = self.kwargs['function_args']['function_data_keys']
                    input_data = [ self.get_data(function_data_key) for function_data_key in function_data_keys ]

                    # Apply the function
                    data = self.kwargs['function_args']['function']( input_data )

                # Other
                else:
                    data = self.data[data_key]

            # Calculate missing data
            except KeyError as e:
                self.handle_data_key_error( data_key )
                continue

            break

        if 'data' not in locals().keys():
            raise KeyError( "After {} tries, unable to find or create data_key, {}".format( i + 1, data_key ) )

        if sl is not None:
            return data[sl]

        return data

    ########################################################################

    def get_position_data( self, data_key ):
        '''Get position data (assuming the data starts with an 'R')

        Args:
            data_key (str) : Key in the data dictionary for the key we want to get

        Returns:
            data (np.ndarray) : Requested data.
        '''

        self.center_coords()

        # Transpose in order to account for when the data isn't regularly shaped
        if data_key == 'Rx':
            data = self.data['P'][0, :]
        elif data_key == 'Ry':
            data = self.data['P'][1, :]
        elif data_key == 'Rz':
            data = self.data['P'][2, :]
        else:
            data = self.data[data_key]

        return data

    ########################################################################

    def get_velocity_data( self, data_key ):
        '''Get position data (assuming the data starts with a 'V')

        Args:
            data_key (str) : Key in the data dictionary for the key we want to get

        Returns:
            data (np.ndarray) : Requested data.
        '''

        self.center_vel_coords()
        self.add_hubble_flow()

        # Get data
        if data_key == 'Vx':
            data = self.data['V'][0, :]
        elif data_key == 'Vy':
            data = self.data['V'][1, :]
        elif data_key == 'Vz':
            data = self.data['V'][2, :]
        else:
            data = self.data[data_key]

        return data

    ########################################################################

    def handle_data_key_error( self, data_key ):
        '''When get_data() fails to data_key in self.data, it passes the data_key to try and generate that data.

        Args:
            data_key (str) : Key to try and generate data for

        Modifies:
            self.data[data_key] (np.array) : If it finds a function to generate the data, it will do so
        '''

        if self.verbose:
            print( 'Data key {} not found in data. Attempting to calculate.'.format( data_key ) )

        method_str = 'calc_{}'.format( data_key )
        if hasattr( self, method_str ):
            getattr( self, method_str )()

        # SimulationData methods
        elif data_key == 'R':
            self.calc_radial_distance()
        elif data_key == 'Vmag':
            self.calc_velocity_magnitude()
        elif data_key == 'Vr':
            self.calc_radial_velocity()
        elif data_key == 'Vtan':
            self.calc_tangential_velocity()
        elif data_key == 'ind':
            self.calc_inds()
        elif data_key == 'L':
            self.calc_ang_momentum()
        elif data_key == 'Phi':
            self.calc_phi()
        elif data_key == 'AbsPhi':
            self.calc_abs_phi()
        elif data_key == 'NumDen':
            self.calc_num_den()
        elif data_key == 'HDen':
            self.calc_H_den()
        elif data_key == 'HIDen':
            self.calc_HI_den()

        # TODO: Move these to the subclasses somehow.
        # Subclass methods
        elif data_key == 'T':
            # TODO: This is a hacky fix, that should be changed...
            if hasattr( self, 'calc_temp' ):
                self.calc_temp()
        elif data_key == 'Pressure':
            self.calc_pressure()

        else:
            raise KeyError( 'NULL data_key, data_key = {}'.format( data_key ) )

    ########################################################################

    def get_distance_to_point( self, point ):
        '''This is *not* unit tested.

        Args:
            point (array-like of shape (3,)) : The point you want to find the distance to for each particle.
        '''

        d_to_point = scipy.spatial.distance.cdist( self.get_data( 'P' ).transpose(), point[np.newaxis, :] )

        return d_to_point.flatten()

    ########################################################################

    def get_potential( self, point ):
        '''This is *not* unit tested.

        Args:
            point (array-like of shape (3,)) : The point you want to find the potential at
        '''

        d_to_point = self.get_distance_to_point( point )

        potential_per_particle = -1. * constants.UNITG_UNIV * self.get_data( 'M' ) / d_to_point

        total_potential = potential_per_particle.sum()

        return total_potential

    ########################################################################
    # Full calculations based on the data
    ########################################################################

    def calc_radial_distance( self ):
        '''Calculate the distance from the origin for a given particle.'''

        self.data['R'] = np.sqrt(
            self.get_data( 'Rx' )**2. + \
            self.get_data( 'Ry' )**2. + \
            self.get_data( 'Rz' )**2.
        )


########################################################################
########################################################################


class SnapshotData( SimulationData ):
    '''Class for analysis of a single snapshot of data.'''

    def __init__( self, snum, *args, **kwargs ):
        '''
        Args:
            snum (int or array of ints) : Snapshot or snapshots to inspect.
        '''

        self.snum = snum

        super( SnapshotData, self ).__init__( *args, **kwargs )

    ########################################################################
    # Get Additional Data
    ########################################################################

    def retrieve_halo_data( self ):

        if self.halo_data_retrieved:
            return

        # Load the AHF data
        ahf_reader = read_ahf.AHFReader( self.halo_data_dir )
        ahf_reader.get_mtree_halos( index=self.ahf_index, tag=self.ahf_tag )

        # Select the main halo at the right redshift
        mtree_halo = ahf_reader.mtree_halos[self.main_halo_id].loc[self.snum]

        # Add the halo data to the class.
        self.redshift = mtree_halo['redshift']
        halo_coords_comoving = np.array( [ mtree_halo['Xc'], mtree_halo['Yc'], mtree_halo['Zc'] ] )
        self.halo_coords = halo_coords_comoving / (1. + self.redshift) / self.data_attrs['hubble']
        self.halo_velocity = np.array( [ mtree_halo['VXc'], mtree_halo['VYc'], mtree_halo['VZc'] ] )
        self.r_vir = mtree_halo['Rvir'] / (1. + self.redshift) / self.data_attrs['hubble']
        self.r_scale = self.r_vir / mtree_halo['cAnalytic']
        self.m_vir = mtree_halo['Mvir'] / self.data_attrs['hubble']
        self.m_gas = mtree_halo['M_gas'] / self.data_attrs['hubble']
        self.m_star = mtree_halo['M_star'] / self.data_attrs['hubble']

        # Calculate the circular velocity
        self.v_c = astro.circular_velocity( self.r_vir, self.m_vir )

        self.halo_data_retrieved = True

        if self.store_ahf_reader:
            self.ahf_reader = ahf_reader

    ########################################################################
    # Properties
    ########################################################################

    @property
    def central_mask( self ):
        '''This mask is used when, for example, finding the velocity of the
        center of mass.
        '''

        if not hasattr( self, '_central_mask' ):

            self._central_mask  = self.data_masker.mask_data(
                'Rf',
                0.,
                self.averaging_frac,
                return_or_store = 'return',
            )

        return self._central_mask

    ########################################################################

    @property
    def v_com( self ):
        '''Property for the velocity of the center of mass.'''

        if not hasattr( self, '_v_com' ):

            m_ma = self.get_selected_data( 'M', self.central_mask )
            v_ma = self.get_selected_data( 'V', self.central_mask )

            self._v_com = ( v_ma * m_ma ).sum( 1 ) / m_ma.sum()

        return self._v_com

    ########################################################################

    @property
    def total_ang_momentum(self):
        '''Calculate the total angular momentum vector.'''

        # Exit early if already calculated.
        if not hasattr( self, '_total_ang_momentum' ):

            # Adapt for application to 'l', which is a multidimensional array
            inner_mask = np.array([ self.central_mask ] * 3)

            # Apply masks
            ang_momentum = self.get_data('L')
            l_ma = np.ma.masked_array(ang_momentum, mask=inner_mask)

            # Get the total angular momentum
            self._total_ang_momentum = np.zeros(3)
            for i in range(3):
                self._total_ang_momentum[i] = l_ma[i].sum()

        return self._total_ang_momentum

    ########################################################################

    @property
    def dN_halo(self, R_vir=None, time_units='abs_length'):
        '''Calculate dN_halo/dX/dlog10Mh or dN_halo/dz/dlog10Mh. X is absorption path length (see for example Ribaudo+11)

        time_units: 'abs_length'- dN_halo/dX/dlog10Mh
                                'redshift' - dN_halo/dz/dlog10Mh
        R_vir:      None - Default, assumes given.
                                'BN' Calculates R_vir from the mass and redshift
        '''

        raise Exception( "TODO: Test this" )

        # Previous code that needs to be cleaned up below.

        # # Choose cosmology
        # cosmo = Cosmology.setCosmology('WMAP9')

        # # Calculate the virial radius, if necessary
        # if R_vir is not None:
        #     if R_vir == 'BN':
        #         # Calculate R_vir off of the cosmocode def, and convert to proper kpc.
        #         self.R_vir = cosmo.virialRadius(M, z) * 10.**3.

        # # Make h easier to use (don't have to write the whole thing out...)
        # h = self.data_attrs['hubble']

        # Mh = h * self.M_vir  # Convert the halo mass to Msun/h, so as to feed it into the HMF.
        # dldz = np.abs(cosmo.line_elt(self.redshift))  # Cosmological line element in Mpc.
        # dldz_kpc = dldz * 10.**3.
        # dndlog10M = cosmo.HMF(Mh, self.redshift) * self.M_vir * np.log(10) * h**4.  # HMF in 1/Mpc^3
        # dndlog10M_kpc = dndlog10M * 10.**-9.
        # dN_halo = dldz_kpc * dndlog10M_kpc * np.pi * self.R_vir**2.

        # # Convert from per redshift to per absorption path length.
        # if time_units == 'abs_length':
        #     dN_halo /= cosmo.dXdz(z)
        # elif time_units == 'redshift':
        #     pass

        # return dN_halo

    ########################################################################
    # Full calculations of the data
    ########################################################################

    def calc_velocity_magnitude(self):
        '''Calculate the radial velocity.'''

        # Center velocity and radius
        self.center_coords()
        self.center_vel_coords()

        # Calculate the radial velocity
        self.data['Vmag'] = np.sqrt(
            self.get_data( 'Vx' )**2. + \
            self.get_data( 'Vy' )**2. + \
            self.get_data( 'Vz' )**2.
        )

    ########################################################################

    def calc_radial_velocity(self):
        '''Calculate the radial velocity.'''

        # Center velocity and radius
        self.center_coords()
        self.center_vel_coords()

        # Calculate the radial velocity
        self.data['Vr'] = ( self.get_data( 'V' ) * self.get_data('P')).sum(0) / self.get_data('R')

    ########################################################################

    def calc_tangential_velocity(self):
        '''Calculate the radial velocity.'''

        # Center velocity and radius
        self.center_coords()
        self.center_vel_coords()

        # Calculate the radial velocity
        self.data['Vtan'] = np.sqrt( self.get_data( 'Vmag' )**2. - self.get_data( 'Vr' )**2. )

    ########################################################################

    def calc_inds(self):
        '''Calculate the indices the data are located at, prior to any masks.'''

        raise Exception( "TODO: Test this" )

        # Flattened index array
        flat_inds = np.arange(self.get_data('Den').size)

        # Put into a multidimensional array
        self.data['ind'] = flat_inds.reshape(self.get_data('Den').shape)

    ########################################################################

    def calc_ang_momentum( self ):
        '''Calculate the angular momentum.'''

        m_mult = np.array( [ self.get_data('M'), ] * 3 )

        p = self.get_data('P')
        v = self.get_data('V')

        l = np.cross( p, v, 0, 0).transpose()
        l *= m_mult

        self.data['L'] = l

    ########################################################################

    def calc_phi( self, normal_vector='total ang momentum' ):
        '''Calculate the angle (in degrees) from some vector.
        By default the vector is the total angular momentum.
        '''

        raise Exception( "TODO: Test this" )

        if vector == 'total ang momentum':
            # Calculate the total angular momentum vector, if it's not calculated yet
            self.normal_vector = self.calc_total_ang_momentum()
        else:
            self.normal_vector = normal_vector

        # Get the dot product
        P = self.get_data('P')
        dot_product = np.zeros(P[0, :].shape)
        for i in range(3):
            dot_product += self.v[i] * P[i, :]

        # Isolate for the cosine
        cos_phi = dot_product / self.get_data('R') / np.linalg.norm(self.v)

        # Get the angle (in degrees)
        self.data['Phi'] = np.arccos(cos_phi) * 180. / np.pi

    ########################################################################

    def calc_abs_phi(self, vector='total gas ang momentum'):
        '''The angle (in degrees) from some vector, but don't mirror
        it around 90 degrees (e.g. 135 -> 45 degrees, 180 -> 0 degrees).'''

        raise Exception( "TODO: Test this" )

        # Get the original Phi
        self.calc_phi(vector)

        self.data['AbsPhi'] = np.where(self.data['Phi'] < 90., self.data['Phi'], np.absolute(self.data['Phi'] - 180.))

    ########################################################################

    def calc_num_den(self):
        '''Calculate the number density (it's just a simple conversion...).'''

        self.data['NumDen'] = self.data['Den'] * constants.UNITDENSITY_IN_NUMDEN

    ########################################################################

    def calc_H_den(self):
        '''Calculate the H density in cgs (cm^-3). Assume the H fraction is ~0.75'''

        # Assume Hydrogen makes up 75% of the gas
        X_H = 0.75

        self.data['HDen'] = X_H * self.data['Den'] * constants.UNITDENSITY_IN_NUMDEN

    ########################################################################

    def calc_HI_den(self):
        '''Calculate the HI density in cgs (cm^-3).'''

        raise Exception( "TODO: Test this" )

        # Assume Hydrogen makes up 75% of the gas
        X_H = 0.75

        # Calculate the hydrogen density
        HDen = X_H * self.data['Den'] * constants.gas_den_to_nb

        self.data['HIDen'] = self.data['nHI'] * HDen

    ########################################################################
    # Non-Altering Calculations
    ########################################################################

    def dist_to_point(self, point, units='default'):
        '''Calculate the distance to a point for all particles.

        point : np array that gives the point
        '''

        raise Exception( "TODO: Test this/update to use scipy cdist" )

        # Calculate the distance to the point
        relative_positions = self.get_data('P').transpose() - np.array(point)
        d_mag = np.linalg.norm(relative_positions, axis=1)

        # Put in different units if necessary
        if units == 'default':
            pass
        elif units == 'h':
            d_mag /= self.get_data('h')
        else:
            raise Exception('Null units, units = {}'.format(units))

        return d_mag

    ########################################################################

    def calc_mu(self):
        '''Calculate the mean molecular weight. '''

        y_helium = self.data['Z_Species'][:, 0]  # Get the mass fraction of helium
        mu = 1. / (1. - 0.75 * y_helium + self.data['ne'])

        return mu

########################################################################
########################################################################

class TimeData( SimulationData ):
    '''Class for analysis of a time series data, e.g. the worldlines of a number of particles.'''

    def __init__(
        self,
        data_masker = None,
        *args, **kwargs
    ):

        if data_masker is None:
            data_masker = TimeDataMasker( self )

        super( TimeData, self ).__init__(
            data_masker = data_masker,
            *args,
            **kwargs
        )

    ########################################################################

    @property
    def n_particles( self ):
        '''Number of snapshots, i.e. data points on the time axis.'''

        if not hasattr( self, '_n_particles' ):
            self._n_particles = self.base_data_shape[0]

        return self._n_particles

    ########################################################################

    @property
    def n_snaps( self ):
        '''Number of snapshots, i.e. data points on the time axis.'''

        if not hasattr( self, '_n_snaps' ):
            self._n_snaps = self.base_data_shape[1]

        return self._n_snaps

    ########################################################################

    def retrieve_halo_data( self ):

        if self.halo_data_retrieved:
            return

        # Load the AHF data
        ahf_reader = read_ahf.AHFReader( self.halo_data_dir )
        ahf_reader.get_mtree_halos( index=self.ahf_index, tag=self.ahf_tag )

        # Select the main halo at the right redshift
        mtree_halo = ahf_reader.mtree_halos[self.main_halo_id].loc[self.snums]

        # Add the halo data to the class.
        self.redshift = mtree_halo['redshift']
        scale_factor_and_hinv = 1. / (1. + self.redshift) / self.hubble_param

        halo_coords_comoving = np.array( [ mtree_halo['Xc'], mtree_halo['Yc'], mtree_halo['Zc'] ] )
        self.halo_coords = halo_coords_comoving * scale_factor_and_hinv[np.newaxis, :]
        self.halo_velocity = np.array( [ mtree_halo['VXc'], mtree_halo['VYc'], mtree_halo['VZc'] ] )
        self.r_vir = mtree_halo['Rvir'] * scale_factor_and_hinv
        self.r_scale = self.r_vir / mtree_halo['cAnalytic']
        self.m_vir = mtree_halo['Mvir'] / self.hubble_param
        self.m_gas = mtree_halo['M_gas'] / self.hubble_param
        self.m_star = mtree_halo['M_star'] / self.hubble_param

        # Calculate the circular velocity
        self.v_c = astro.circular_velocity( self.r_vir, self.m_vir )

        self.halo_data_retrieved = True

        if self.store_ahf_reader:
            self.ahf_reader = ahf_reader

    ########################################################################

    def calc_radial_velocity( self ):
        '''Calculate the distance from the origin for a given particle.'''

        v_all = self.get_data( 'V' )
        p_all = self.get_data( 'P' )
        r_all = self.get_data( 'R' )

        # Note that we don't need to add the hubble flow, because we've
        # already done so in self.get_velocity_data()
        assert self.hubble_corrected

        v_r = []
        for i in range( self.n_snaps ):

            v = v_all[:,:,i]
            p = p_all[:,:,i]
            r = r_all[:,i]

            v_r_i = ( v * p ).sum( axis=0 )/r

            v_r.append( v_r_i )

        # Format the output
        self.data['Vr'] = np.array( v_r ).transpose()

    ########################################################################

    def handle_data_key_error( self, data_key ):

        method_str = 'calc_{}'.format( data_key )
        if hasattr( self, method_str ):
            getattr( self, method_str )()
        elif self.calc_time_as_classification( data_key ):
            return

        else:
            super( TimeData, self ).handle_data_key_error( data_key )
            # raise KeyError( 'NULL data_key, data_key = {}'.format( data_key ) )

    ########################################################################

    def get_processed_data(
        self,
        data_key,
        sl = None,
        smooth_data = False,
        smoothing_window_length = 9,
        smoothing_polyorder = 3,
        a_power = None,
        scale_key = None,
        scale_a_power = None,
        scale_h_power = None,
        tile_data = False,
        tile_dim = 'auto',
        *args, **kwargs
    ):
        '''Modified method for getting processed method. For the most part is
        equivalent to calling the method of the parent class, but is also
        capable of scaling the retrieved data by a column from the halo data.

        Args:
            data_key (str) : What to get out?

            sl (slice) : Slice of the data, if requested.

            smooth_data (bool) : If True, smooth the data.

            smoothing_window_length, smoothing_polyorder (int) :
                Arguments for how to smooth the data.

            scale_key (str) :
                Halo data entry by which to divide the data by.

            scale_a_power (float) :
                The halo data that we are scaling processed_data by will be
                multiplied by a to this power.
                Useful for data in cosmological units (as is often normal).

            scale_h_power (float) :
                The halo data that we are scaling processed_data by will be
                multiplied by the hubble parameter to this power.
                Useful for data in cosmological units (as is often normal).

            tile_data (bool) :
                If True, tile data along a given direction. This is usually for
                data formatting purposes.

            tile_dim (str) :
                If the data is tiled, what dimension of the data should match?
                Options:
                    'auto' :
                        Tiles according to data size.
                    'match_snaps' :
                        The data is tiled such that the new shape is
                        (self.n_snaps, data_size).
                    'match_particles' :
                        The data is tiled such that the new shape is
                        (data_size, self.n_particles).

            *args, **kwargs :
                Passed to SimulationData.get_processed_data.

        Returns:
            processed_data (array-like) : Requested data array.
        '''

        if ( sl is not None ) and tile_data:
            used_sl = None
        else:
            used_sl = sl

        processed_data = super( TimeData, self ).get_processed_data(
            data_key,
            sl = used_sl,
            *args, **kwargs
        )

        if smooth_data:
            processed_data = signal.savgol_filter(
                processed_data,
                window_length = smoothing_window_length,
                polyorder = smoothing_polyorder,
            )

        if scale_key is not None:

            # Get the data
            data_to_div_by = self.halo_data.get_mt_data(
                scale_key,
                mt_halo_id = self.main_halo_id,
                a_power = scale_a_power,
                snums = self.snums
            )

            if scale_h_power is not None:
                processed_data /= self.hubble_param**scale_h_power

            if sl is not None:
                # DEBUG
                try:
                    data_to_div_by = data_to_div_by[sl]
                # For when we're getting a slice of a single snapshot
                except IndexError:
                    assert sl[0] == slice(None)
                    data_to_div_by = data_to_div_by[sl[1]]

            processed_data /= data_to_div_by

        if tile_data:

            if tile_dim == 'auto':
                if processed_data.shape == ( self.n_particles, ):
                    tile_dim = 'match_snaps'
                elif processed_data.shape == ( self.n_snaps, ):
                    tile_dim = 'match_particles'
                elif processed_data.shape == self.base_data_shape:
                    tile_dim = None
                else:
                    raise Exception(
                        "Unrecognized data shape, {}".format(
                            processed_data.shape
                        )
                    )

            if tile_dim == 'match_snaps':
                processed_data = np.tile(
                    processed_data,
                    ( self.n_snaps, 1),
                ).transpose()

            elif tile_dim == 'match_particles':
                processed_data = np.tile(
                    processed_data,
                    ( self.n_particles, 1),
                )

            if sl is not None:
                processed_data = processed_data[sl]

        return processed_data

    ########################################################################

    def get_selected_data_over_time( self, *args, **kwargs ):
        '''Wrapper for geting masked data as a function of time.'''

        return self.data_masker.get_selected_data_over_time( *args, **kwargs )

    ########################################################################

    @property
    def snums( self ):

        # TODO: This is a workable structure for now, but it's not ideal. This may not always be how we get the snum
        return self.data['snum']

    ########################################################################

    @property
    def hubble_param( self ):

        # TODO: This is a workable structure for now, but it's not ideal. The hubble parameter may not always be here.
        return self.data_attrs['hubble']

    ########################################################################

    def calc_ang_momentum( self ):
        '''The angular momentum (in the standard coordinates).

        Modifies:
            self.data['L'] (array-like):
                Angular momentum of each resolution element.
        '''

        m_mult = np.array( [ self.get_data('M'), ] * 3 )

        p_all = self.get_data('P')
        v_all = self.get_data('V')

        l_all = np.zeros( ( 3, self.n_particles, self.n_snaps ) )
        # Calculate the angular momentum at each redshift
        for i in range( self.n_snaps ):                                         
                                                                                
            v = v_all[:,:,i]                                                    
            p = p_all[:,:,i]                                                    
            m = m_mult[:,:,i] 

            l = np.cross( p, v, 0, 0).transpose()
            l *= m

            l_all[:,:,i] = l

        self.data['L'] = l_all

    ########################################################################

    def calc_phi( self, normal_vector='total ang momentum' ):
        '''Calculate the angle (in degrees) from some vector.
        By default the vector is the total angular momentum.

        Args:
            normal_vector (str or array-like):
                Vector that represents the vertical.
        '''

        # Exit early if already calculated with the same normal vector
        if 'Phi' in self.data:
            same = np.allclose(
                self.normal_vector,
                normal_vector
            )
            if same:
                return

        # Set up the normal vector
        if normal_vector == 'total ang momentum':
            self.normal_vector = self.total_ang_momentum
        else:
            self.normal_vector = normal_vector

        # Get all the data we need
        p_all = self.get_data('P')
        r_all = self.get_data('R')

        # Loop over each snapshot and calculate
        phi_all = np.zeros( self.base_data_shape )
        for i in range( self.n_snaps ):

            p = p_all[:,:,i]                                                    
            r = r_all[:,i]

            # Get the dot product
            dot_product = np.zeros( p[0,:].shape )
            for j in range(3):
                dot_product += self.normal_vector[j] * p[j,:]

            # Isolate for the cosine
            cos_phi = dot_product / r / np.linalg.norm( self.normal_vector )

            # Get the angle (in degrees)
            phi_all[:,i] = np.arccos( cos_phi ) * 180. / np.pi

        self.data['Phi'] = phi_all

    ########################################################################

    def calc_abs_phi(self, normal_vector='total ang momentum' ):
        '''Calculate the angle (in degrees) from some vector, but mirror
        values past 90 degrees (e.g. 135 -> 45 degrees, 180 -> 0 degrees).
        This is useful when there's symmetry above and below 90.

        Args:
            normal_vector (str or array-like):
                Vector that represents the vertical. Defaults to using the
                total stellar angular momentum of the main galaxy.

        Modifies:
            self.data['AbsPhi'] (array-like):
                Angle from the vector, in many cases acting as the angle
                from the disk axis.
        '''

        # Get the original Phi
        self.calc_phi( normal_vector )

        phi = self.data['Phi']

        # Actual calculation
        self.data['AbsPhi'] = np.where(
            phi < 90.,
            phi,
            np.absolute( phi - 180.)
        )

    ########################################################################

    def calc_time_as_classification( self, data_key ):

        # Check if we should be running this function (does the provided
        # data_key even match the format we want to parse?)
        if data_key[:7] != 'time_as':
            return False

        # Get the data key for the classification
        classification_data_key = 'is_{}'.format( data_key[8:] )

        # Get out the classification data itself
        classification = self.get_data( classification_data_key )

        # Get out the time intervals, and tile them for formatting
        dt = self.get_data( 'dt' )

        # Fill in the array row by row
        time_as_classification = np.zeros( classification.shape )
        for i, row in enumerate( classification ):

            # Identify regions of contiguous classification
            contiguous_regions = data_operations.contiguous_regions( row )

            # Find the cumulative time in specified regions
            for start, end in contiguous_regions:

                dt_region = dt[start:end]

                # We need to flip when we sum because we want a reverse sum
                # (due to data formatting, t=0 is at j=-1)
                cumtime_region = np.cumsum( dt_region[::-1] )[::-1]

                # Store that time
                time_as_classification[i, start:end] = cumtime_region

        self.data[data_key] = time_as_classification

        return True

    ########################################################################

    def calc_metal_mass( self ):
        '''Calculate the metal mass held by each resolution element.

        Modifies:
            self.data['enriched_metal_mass'] (array-like):
                Metal mass from enrichment.
        '''

        metal_mass = self.get_data( 'M' ) * self.get_data( 'Z' ) * self.z_sun

        self.data['metal_mass'] = metal_mass

    ########################################################################

    def calc_enriched_metal_mass( self ):
        '''Calculate the metal mass that comes from enrichment for
        each resolution element, not counting mass that's at the metallicity
        floor. Assumes that the there will always be at least one resolution
        element in the simulation that's at the metallicity floor.

        Modifies:
            self.data['enriched_metal_mass'] (array-like):
                Metal mass from enrichment.
        '''

        enrichment_fraction = (
            self.get_data( 'Z' ) - np.nanmin( self.get_data( 'Z' ) )
        )

        enrichment_fraction *= self.z_sun

        enriched_metal_mass = self.get_data( 'M' ) * enrichment_fraction

        self.data['enriched_metal_mass'] = enriched_metal_mass

########################################################################
########################################################################

class TimeDataMasker( generic_data.DataMasker ):
    '''Data masker for worldline data.'''

    def __init__( self, time_data ):

        super( TimeDataMasker, self ).__init__( time_data )

    ########################################################################

    def get_selected_data_over_time(
        self,
        data_key,
        snum,
        mask = 'total',
        optional_masks = None,
        n_samples = None,
        seed = None,
        *args, **kwargs
    ):
        '''Get data over the full time history, based on its mask at
        one time.
    
        Args:
            data_key (str): Data to get.
            
            snum (int): Snapshot to get the data corresponding to.

            mask (
        '''

        # Make sure we don't try to pass a slice to any keyword arguments
        assert 'sl' not in kwargs, 'Taking slices of the original data' + \
                ' should not be done when using get_selected_data_over_time'

        # Get the appropriate mask
        if isinstance( mask, np.ndarray ):
            used_mask = mask
        elif isinstance( mask, bool ) or isinstance( mask, np.bool_ ):
            if not mask:
                return self.data_object.get_processed_data(
                    data_key,
                    *args, **kwargs
                )
            
            raise Exception( "All data is masked." )
        elif mask == 'total':
            used_mask = self.get_total_mask( optional_masks=optional_masks )
        else:
            raise KeyError( "Unrecognized type of mask, {}".format( mask ) )

        # Get the relevant ind
        correct_snum = self.data_object.snums == snum
        ind = np.arange( self.data_object.n_snaps )[correct_snum][0]

        # Get the boolean for the selected data
        sl = np.invert( used_mask[:,ind] )

        # If we want to get the particle ind, we do something different
        if data_key == 'particle_ind':
            if n_samples is None:
                n_particles_selected = sl.sum()
            else:
                n_particles_selected = n_samples

            return np.tile(
                np.arange( n_particles_selected ),
                ( self.data_object.n_snaps, 1 ),
            ).transpose()
        
        # Get the masked data (masked via the slice)
        masked_data = self.data_object.get_processed_data(
            data_key,
            sl = sl,
            *args, **kwargs
        )

        # Sample a subset of the data
        if n_samples is not None:

            # If given a seed for sampling, use it
            if seed is not None:
                np.random.seed( seed )

            n_selected_particles = masked_data.shape[0]
            if n_samples >= n_selected_particles:
                print( "n_samples > n_selected_particles, not sampling." )
                return masked_data

            try:
                sampled_inds = np.random.choice(
                    np.arange( n_selected_particles ),
                    n_samples,
                    replace = False,
                )
                masked_data = masked_data[sampled_inds,:]
            except:
                #DEBUG
                import pdb; pdb.set_trace()

        return masked_data








#!/usr/bin/env python
'''General purpose plotting tools.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import copy
import errno
import numpy as np
import os
import subprocess

import matplotlib.pyplot as plt
import matplotlib.colors as plt_colors
import matplotlib.patches
from mpl_toolkits.axes_grid1 import make_axes_locatable

import galaxy_dive.utils.data_operations as data_ops
import galaxy_dive.utils.astro as astro_utils

########################################################################
########################################################################

def box_plot(
    x_datas,
    y_datas,
    ax = None,
    color = 'k',
    x_scale = 'log',
    y_scale = 'log',
    y_floor = None,
    box_upper_limit = None,
    box_lower_limit = None,
    box_zorder = 50,
    blank_zorder = 30,
    line_zorder = 10,
    zorder_offset = 0,
    plot_boxes = True,
    skip_single_points = True,
    linewidth = 4,
):

    if ax is None:
        fig = plt.figure( figsize=(11,6) )
        ax = plt.gca()

    means = {
        'x' : [],
        'y' : [],
    }
    errs = {
        'x' : [],
        'y' : [],
    }
    for x_data, y_data in zip( x_datas, y_datas ):

        if ( len( x_data ) < 2 ) and skip_single_points:
            continue

        # Make copies so we don't modify the data
        x_data = copy.copy( x_data )
        y_data = copy.copy( y_data )

        # Apply a floor
        y_data[y_data<y_floor] = y_floor

        # Put in logspace
        if x_scale == 'log':
            x_data = np.log10( x_data )
        if y_scale == 'log':
            y_data = np.log10( y_data )

        x_mean = x_data.mean()
        y_mean = y_data.mean()

        x_err = x_data.std()
        y_err = y_data.std()

        # Get the negative and positive errs
        if x_scale == 'log':
            x_err_neg, x_err_pos = errs_from_log_errs(
                x_mean,
                x_err,
            )
        elif x_scale == 'linear':
            x_err_neg = x_err
            x_err_pos = x_err
        if y_scale == 'log':
            y_err_neg, y_err_pos = errs_from_log_errs(
                y_mean,
                y_err,
                upper_limit = box_upper_limit,
                lower_limit = box_lower_limit,
            )
        elif y_scale == 'linear':
            y_err_neg = y_err
            y_err_pos = y_err

        if x_scale == 'log':
            x_mean = 10.**x_mean
        if y_scale == 'log':
            y_mean = 10.**y_mean

        # Store for plotting
        means['x'].append( x_mean )
        means['y'].append( y_mean )
        errs['x'].append( x_mean )
        errs['y'].append( x_mean )

        left = x_mean - x_err_neg
        bottom = y_mean - y_err_neg

        width = x_err_neg + x_err_pos
        height = y_err_neg + y_err_pos

        if plot_boxes:
            # Rectangle itself
            rect = matplotlib.patches.Rectangle(
                (left,bottom),
                width,
                height,
                fill = False,
                edgecolor = color,
                linewidth = linewidth,
                zorder = box_zorder - zorder_offset,
            )                
            ax.add_patch( rect )
            
            # Blank rectangles to clean things up
            rect = matplotlib.patches.Rectangle(
                (left,bottom),
                width,
                height,
                facecolor = 'w',
                edgecolor = 'w',
                linewidth = linewidth,
                zorder = blank_zorder - zorder_offset,
            )                
            ax.add_patch( rect )

    sorted_inds = np.argsort( means['x'] )

    # Plot everything else
    ax.plot(
        np.array( means['x'] )[sorted_inds],
        np.array( means['y'] )[sorted_inds],
        linewidth = linewidth,
        color = color,
        zorder = line_zorder - zorder_offset,
    )

    # Set scale
    ax.set_xscale( x_scale )
    ax.set_yscale( y_scale )

########################################################################

def errs_from_log_errs( log_mean, log_err, upper_limit=None, lower_limit=None ):
    
    err_positive = ( 10.**log_mean ) * \
        ( 10.**log_err - 1. )
    err_negative = -1.*( 10.**log_mean ) * \
        ( 10.**-log_err - 1. )
        
    if upper_limit is not None:
        exceeds_limit = 10.**log_mean + err_positive > upper_limit
        try:
            err_positive[exceeds_limit] = ( upper_limit - 10.**log_mean )[exceeds_limit]
        except TypeError:
            err_positive = ( upper_limit - 10.**log_mean )

    if lower_limit is not None:
        exceeds_limit = 10.**log_mean - err_negative < lower_limit
        try:
            err_negative[exceeds_limit] = ( 10.**log_mean - lower_limit )[exceeds_limit]
        except TypeError:
            err_negative = ( 10.**log_mean - lower_limit )
        
    return err_negative, err_positive

########################################################################

def add_redshift_to_axis(
        ax,
        hubble,
        omega_matter,
        tick_redshifts = np.array(
            [ 0.1, 0.25, 0.4, 0.5, 0.75, 1, 1.5, 2, 2.5, 3, 4, 5, ]
        )
    ):
    '''Add redshift labels to the top axis of plot that has age of the universe
    on the bottom axis.

    Args:
        ax (axis object): Original axis

        hubble (float): Hubble parameter, used for calculating the age.

        omega_matter (float): Cosmological param used for calcing the age.

        tick_redshifts (array-like): Redshifts to plot ticks at.
    
    '''

    tick_times = astro_utils.age_of_universe(
        tick_redshifts,
        h = hubble,
        omega_matter = omega_matter,
    )

    # Make sure we aren't trying to plot ticks that would go out of bounds,
    # because that breaks things
    ax2_ticks = []
    ax2_tick_labels = []
    x_range = ax.get_xlim()
    for ax2_tick, ax2_tick_label in zip( tick_times, tick_redshifts ):
        if ( ax2_tick > x_range[0] ) and ( ax2_tick < x_range[1] ):
            ax2_ticks.append( ax2_tick )
            ax2_tick_labels.append( ax2_tick_label )

    # Add a second axis for plotting
    ax2 = ax.twiny()
    ax2.set_xlim( x_range )
    ax2.set_xticks( ax2_ticks )
    ax2.set_xticklabels( ax2_tick_labels )

    return ax2
        
########################################################################

def make_movie( img_dir, file_pattern, movie_save_file, quality=1 ):
    '''Make a movie using ffmpeg.

    Args:
        img_dir (str) : Directory containing images to stitch together.
        file_pattern (str) : Pattern for the images.
        movie_save_file (str) : What to save the movie as. Typically .mp4 or .avi
        quality (int) : What quality to use for making the movie (1-31). Default is 1, which is highest quality.
    '''

    prev_dir = os.getcwd()
    os.chdir( img_dir )

    subprocess.call( [ "ffmpeg", "-pattern_type", "glob", "-i", file_pattern, "-q:v", str( quality ), movie_save_file, ])

    os.chdir( prev_dir )

########################################################################


def add_colorbar( fig_or_ax, color_object, method='fig', ax_location=[0.9, 0.1, 0.03, 0.8], **kw_args ):
    '''Add a colorbar to a figure(fig) according to an object on the figure(color_object). Added to the right side of the figure.
    Or add a colorbar to the right of an axis.

    Args:
        fig_or_ax (figure or axis object): The object to add the colorbar to.
        color_object (object): The object providing the colorscheme, i.e. what you want the colorbar to provide a map for.
        method (str): If 'fig', add to the right of the figure. If 'ax', add to the right of the axis.
        ax_location (list): If adding to a figure, this is where and how large the axis will be.
        **kw_args will be passed to colorbar()
    '''

    if method == 'fig':

        cax = fig_or_ax.add_axes(ax_location)
        cbar = fig_or_ax.colorbar(color_object, cax=cax, **kw_args)

    if method == 'ax':

        # Create divider for existing axes instance
        divider = make_axes_locatable(fig_or_ax)

        # Append axes to the right of ax, with 5% width of ax
        cax = divider.append_axes("right", pad=0.05, size='5%')

        # Create colorbar in the appended axes
        cbar = plt.colorbar(color_object, cax=cax, **kw_args)

    return cbar

########################################################################

# Functions for switching between a single index and a pair of indexes in the same way that add_subplot maps single subplots.


def custom_sequential_colormap(
    color1,
    color2 = (1, 1, 1),
    cmap_name = 'custom',
    n_bins = 256,
):
    '''Create a custom sequential colormap using one to two colors

    Args:
        color1 (color identifier, e.g. hex or RGB tuple) :
            One of the two colors used. The colormap transtions from color1
            to color2 on a gradient.

        color2 (color identifier, e.g. hex or RGB tuple) :
            One of the two colors used. The colormap transtions from color1
            to color2 on a gradient.

        cmap_name (str) :
            What to call the colormap.

        n_bins (int) :
            Number of bins to divide the colormap into.

    Returns:
        cmap (colormap instance) :
            Resulting colormap.
    '''

    colors = [ color1, color2 ]

    cmap = plt_colors.LinearSegmentedColormap.from_list(
        cmap_name,
        colors,
        N = n_bins
    )

    return cmap


def reading_indexes( index, num_cols ):
    '''Switch to multiple indexes for a plot.

    Args:
        index (int): An index that can be used to access a subplot in the same way add_subplot() accesses subplots.
        num_cols (int): Number of columns in the figure.

    Returns:
        horizontal_index (int): Horizontal index of the plot.
        vertical_index (int): Vertical index of the plot.
    '''

    horizontal_index = index % num_cols
    vertical_index = index / num_cols

    return (horizontal_index, vertical_index)


def single_index( horizontal_index, vertical_index, num_cols ):
    '''Switch to a single index for a plot.

    Args:
        horizontal_index (int): Horizontal index of the plot.
        vertical_index (int): Vertical index of the plot.
        num_cols (int): Number of columns in the figure.

    Returns:
        index (int): An index that can be used to access a subplot in the same way add_subplot() accesses subplots.
    '''

    return vertical_index * num_cols + horizontal_index

########################################################################


def calc_num_rows( num_plots, num_cols ):
    '''Find the number of plots given the number of plots and the number of columns.

    Args:
        num_plots (int): Number of subplots.
        num_cols (int): Number of columns in the figure.

    Returns:
        num_rows (int): Number of rows in the figure.
    '''

    h_index_end, v_index_end = reading_indexes(num_plots - 1, num_cols)
    num_rows = v_index_end + 1

    return num_rows

########################################################################


def remove_all_space( figure, num_rows, num_cols, num_plots ):
    '''Removes any space between subplots for the given figure instance.

    Args:
        figure (figure object): The figure to alter.
        num_rows (int): Number of rows in the figure.
        num_cols (int): Number of columns in the figure.
        num_plots (int): Number of subplots.

    Modifies:
        figure (figure object): Alters the spacing.
    '''

    # Remove any space between the plots.
    figure.subplots_adjust(hspace=0.0001, wspace=0.0001)

    for k in range(0, num_plots):

        # Get the subplot to plot it on.
        (i, j) = reading_indexes(k, num_cols)
        ax = figure.add_subplot(num_rows, num_cols, k + 1)

        # Get rid of x tick marks for middle graphs
        if single_index(i, j + 1, num_cols) < num_plots:

            plt.setp(ax.get_xticklabels(), visible=False)

        # Get rid of y tick marks for middle graphs
        if i > 0:

            plt.setp(ax.get_yticklabels(), visible=False)

        # Get rid of tick labels that overlap with each other
        if i != num_cols - 1 and j == num_rows - 1:

            plt.setp(ax.get_xticklabels()[-1], visible=False)
        if j != 0:

            plt.setp(ax.get_yticklabels()[-1], visible=False)

        # Get rid of x-tick labels that overlap with the other plots.
        if j + 1 == num_rows:

            plt.setp(ax.get_xticklabels()[0], visible=False)

        # Avoid overlapping x- and y-tick marks of the same subplot.

########################################################################


def save_fig( out_dir, save_file, fig=None, resolution='auto', **save_args ):
    '''Save a figure using a pretty frequent combination of details.

    Args:
        out_dir (str): Output directory.
        save_file (str): Save file name.

    Keyword Args:
        fig : Figure or axis.
        resolution : What resolution to save the figure as. Defaults to 3600 pixels on the shortest side.
    '''

    # Make sure the output directory exists
    try:
        os.makedirs(out_dir)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(out_dir):
            pass
        else:
            raise
    
    if resolution == 'auto':
        figsize = np.min( fig.get_size_inches() )
        dpi = int( 3600./figsize )
    else:
        dpi = resolution

    default_save_args = {'dpi': dpi, 'bbox_inches': 'tight'}
    used_save_args = data_ops.merge_dict( save_args, default_save_args)

    # Save the figure.
    if fig is None:
        plt.savefig('%s/%s' % (out_dir, save_file), **used_save_args)
    else:
        fig.savefig('%s/%s' % (out_dir, save_file), **used_save_args)

    print('File saved at %s/%s' % (out_dir, save_file))

########################################################################


def fill_between_steps( ax, x, y1, y2=0, step_where='pre', **kwargs ):
        '''Fill between for a step plot.

        Args:
            ax (Axes) The axes to draw to
            x (array-like): Array/vector of index values.
            y1 (array-like or float): Array/vector of values to be filled under.
            y2 (array-Like or float, optional): Array/vector or bottom values for filled area. Default is 0.
            step_where ({'pre', 'post', 'mid'}): where the step happens, same meanings as for `step`
            **kwargs will be passed to the matplotlib fill_between() function.

        Returns
            ret (PolyCollection): The added artist
        '''
        if step_where not in {'pre', 'post', 'mid'}:
                raise ValueError(
                    "where must be one of {{'pre', 'post', 'mid'}} "
                    "You passed in {wh}".format(wh=step_where))

        # make sure y values are up-converted to arrays
        if np.isscalar(y1):
                y1 = np.ones_like(x) * y1

        if np.isscalar(y2):
                y2 = np.ones_like(x) * y2

        # temporary array for up-converting the values to step corners
        # 3 x 2N - 1 array

        vertices = np.vstack((x, y1, y2))

        # this logic is lifted from lines.py
        # this should probably be centralized someplace
        if step_where == 'pre':
                steps = np.zeros((3, 2 * len(x) - 1), np.float)
                steps[0, 0::2], steps[0, 1::2] = vertices[0, :], vertices[0, :-1]
                steps[1:, 0::2], steps[1:, 1:-1:2] = vertices[1:, :], vertices[1:, 1:]

        elif step_where == 'post':
                steps = np.zeros((3, 2 * len(x) - 1), np.float)
                steps[0, ::2], steps[0, 1:-1:2] = vertices[0, :], vertices[0, 1:]
                steps[1:, 0::2], steps[1:, 1::2] = vertices[1:, :], vertices[1:, :-1]

        elif step_where == 'mid':
                steps = np.zeros((3, 2 * len(x)), np.float)
                steps[0, 1:-1:2] = 0.5 * (vertices[0, :-1] + vertices[0, 1:])
                steps[0, 2::2] = 0.5 * (vertices[0, :-1] + vertices[0, 1:])
                steps[0, 0] = vertices[0, 0]
                steps[0, -1] = vertices[0, -1]
                steps[1:, 0::2], steps[1:, 1::2] = vertices[1:, :], vertices[1:, :]
        else:
                raise RuntimeError("should never hit end of if-elif block for validated input")

        # un-pack
        xx, yy1, yy2 = steps

        # now to the plotting part:
        return ax.fill_between(xx, yy1, y2=yy2, **kwargs)



#!/usr/bin/env python
'''General utilities

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import collections
from contextlib import contextmanager
import errno
from functools import wraps
import importlib
import inspect
import itertools
import numpy as np
import os
import shutil
import subprocess
import sys
import time
# Work for py2 and py3
try:
    from StringIO import StringIO
except ImportError:
        from io import StringIO

########################################################################
########################################################################

class SmartDict( collections.Mapping ):
    '''Replacement for dictionary that allows easier access to the attributes and methods of the dictionary components.
    For example, if one has a smart dictionary of TestClassA objects, each of which has a TestClassB attribute, which
    in turn have a foo method, then smart_dict.test_class_b.foo(2) would be a dict with foo calculated for each.
    In other words, it would be equivalent to the following code:
    results = {}
    for key in smart_dict.keys():
        results[key] = smart_dict[key].test_class_b.foo( 2 )
    return results

    NOTE: In Python 3, the parent class probably needs to be switched to collections.abc.Mapping.
    '''

    def __init__( self, *args, **kwargs ):
        self._storage = dict( *args, **kwargs )

    def __iter__( self ):
        return iter( self._storage )

    def __len__( self ):
        return len( self._storage )

    def __getitem__( self, item ):
        return self._storage[item]

    def __setitem__( self, key, item ):
        self._storage[key] = item

    def __repr__( self ):

        out_str = "SmartDict, {\n"

        for key in self.keys():

            def get_print_obj( obj ):
                try:
                    print_obj = obj.__repr__()
                except:
                    print_obj = obj
                return print_obj

            out_str += "{} : {},\n".format( get_print_obj( key ), get_print_obj( self._storage[key] ), )

        out_str += "}\n"

        return out_str

    def __getattr__( self, attr ):

        results = {}
        for key in self.keys():

            results[key] = getattr( self._storage[key], attr )

        return SmartDict( results )

    def __call__( self, *args, **kwargs ):

        results = {}
        for key in self.keys():

            results[key] = self._storage[key]( *args, **kwargs )

        return SmartDict( results )

    def call_custom_kwargs( self, kwargs, default_kwargs={}, verbose=False ):
        '''Perform call, but using custom keyword arguments per dictionary tag.

        Args:
            kwargs (dict) : Custom keyword arguments to pass.
            default_kwargs (dict) : Defaults shared between keyword arguments.

        Returns:
            results (dict) : Dictionary of results.
        '''

        used_kwargs = dict_from_defaults_and_variations( default_kwargs, kwargs )

        results = {}
        for key in self.keys():

            if verbose:
                print( "Calling for {}".format( key ) )

            results[key] = self._storage[key]( **used_kwargs[key] )

        return SmartDict( results )

    ########################################################################

    def call_iteratively( self, args_list ):

        results = {}
        for key in self.keys():

            inner_results = []
            for args in args_list:

                inner_result = self._storage[key]( args )
                inner_results.append( inner_result )

            results[key] = inner_results

        return results

    ########################################################################
    # For handling when the Smart Dict contains dictionaries
    ########################################################################

    def inner_item( self, item ):
        '''When SmartDict is a dictionary of dicts themselves, this can be used to get an item from those dictionaries.'''

        results = {}

        for key in self.keys():

            results[key] = self._storage[key][item]

        return SmartDict( results )

    def inner_keys( self ):

        results = {}

        for key in self.keys():

            results[key] = self._storage[key].keys()

        return SmartDict( results )

    def transpose( self ):

        results = {}

        # Populate results dictionary
        for key, item in self.items():
            for inner_key in item.keys():

                try:
                    results[inner_key][key] = item[inner_key]
                except KeyError:
                    results[inner_key] = {}
                    results[inner_key][key] = item[inner_key]

        return SmartDict( results )

    ########################################################################
    # Operation Methods
    ########################################################################

    def __add__( self, other ):

        results = {}

        if isinstance( other, SmartDict ):
            for key in self.keys():
                results[key] = self._storage[key] + other[key]

        else:
            for key in self.keys():
                results[key] = self._storage[key] + other

        return SmartDict( results )

    __radd__ = __add__

    def __sub__( self, other ):

        results = {}

        if isinstance( other, SmartDict ):
            for key in self.keys():
                results[key] = self._storage[key] - other[key]

        else:
            for key in self.keys():
                results[key] = self._storage[key] - other

        return SmartDict( results )

    def __rsub__( self, other ):
        results = {}

        if isinstance( other, SmartDict ):
            for key in self.keys():
                results[key] = other[key] - self._storage[key]

        else:
            for key in self.keys():
                results[key] = other - self._storage[key]

        return SmartDict( results )

    def __mul__( self, other ):

        results = {}

        if isinstance( other, SmartDict ):
            for key in self.keys():
                results[key] = self._storage[key]*other[key]

        else:
            for key in self.keys():
                results[key] = self._storage[key]*other

        return SmartDict( results )

    __rmul__ = __mul__

    def __div__( self, other ):

        results = {}

        if isinstance( other, SmartDict ):
            for key in self.keys():
                results[key] = self._storage[key]/other[key]
        else:
            for key in self.keys():
                results[key] = self._storage[key]/other

        return SmartDict( results )

    def __rdiv__( self, other ):

        results = {}

        if isinstance( other, SmartDict ):
            for key in self.keys():
                results[key] = other[key]/self._storage[key]
        else:
            for key in self.keys():
                results[key] = other/self._storage[key]

        return SmartDict( results )

    ########################################################################
    # Other operations
    ########################################################################

    def sum_contents( self ):
        '''Get the sum of all the contents inside the Dict.'''

        for i, key in enumerate( self.keys() ):

            # To handle non-standard items (e.g. things that aren't ints or floats )
            if i == 0:
                result = self._storage[key]

            else:
                result += self._storage[key]

        return result

    def keymax( self ):
        
        for key, item in self.items():
            try:
                if item_max < item:
                    item_max = item
                    key_max = key
            except NameError:
                item_max = item
                key_max = key

        return key_max, item_max

    def keymin( self ):
        
        for key, item in self.items():
            try:
                if item_min > item:
                    item_min = item
                    key_min = key
            except NameError:
                item_min = item
                key_min = key

        return key_min, item_min

    def apply( self, fn, *args, **kwargs ):
        '''Apply some function to each item in the smart dictionary, and
        return the results as a SmartDict.
        '''

        results = {}

        for key, item in self.items():
            results[key] = fn( item, *args, **kwargs )

        return SmartDict( results )

    def split_by_key_slice( self, sl, str_to_match=None ):
        '''Break the smart dictionary into smaller smart dictionaries according
        to a subset of the key.

        Args:
            sl (slice) :
                Part of the key to use to make subsets.

            str_to_match (str) :
                If True, split into broad categories that match this string
                in the given slice or don't.
        '''

        results = {}

        if str_to_match is not None:
            results[True] = SmartDict( {} )
            results[False] = SmartDict( {} )

        for key, item in self.items():

            key_slice = key[sl]

            if str_to_match is not None:
                str_matches = key_slice == str_to_match
                results[str_matches][key] = item
            
            else:
                try:
                    results[key_slice][key] = item
                except KeyError:
                    results[key_slice] = SmartDict( {} )
                    results[key_slice][key] = item

        return results

    def split_by_dict( self, d, return_list=False ):
        '''Break the smart dictionary into smaller smart dictionaries according
        to their label provided by a dictionary

        Args:
            d (dict) :
                Dictionary to use to split into smaller dictionaries

            return_list (bool) :
                If True, return a list of arrays.
        '''

        results = {}
    
        for key, item in self.items():

            try:
                result_subkey = d[key]
                try:
                    results[result_subkey][key] = item
                except KeyError:
                    results[result_subkey] = SmartDict( {} )
                    results[result_subkey][key] = item
            except KeyError:
                try:
                    results['no label'][key] = item
                except KeyError:
                    results['no label'] = SmartDict( {} )
                    results['no label'][key] = item
        
        if not return_list:
            return results

        final_results = []
        for key, item in results.items():
            final_results.append( item.array() )

        return final_results

    def log10( self ):
        '''Wrapper for np.log10'''

        return self.apply( np.log10 )

    def array( self ):
        '''Returns a np.ndarray with unique order (sorted keys )'''

        values = [ x for _,x in sorted(zip( self.keys(), self.values() ) )]

        return np.array( values )

    def remove_empty_items( self ):
        '''Look for empty items and delete them.'''

        keys_to_delete = []
        for key, item in self.items():
            if len( item ) == 0:
                keys_to_delete.append( key )

        for key in keys_to_delete:
            del self._storage[key]

    ########################################################################
    # Construction Methods
    ########################################################################

    @classmethod
    def from_class_and_args( cls, contained_cls, args, default_args={}, ):
        '''Alternate constructor. Creates a SmartDict of contained_cls objects, with arguments passed to it from
        the dictionary created by defaults and variations.

        Args:
            contained_cls (type of object/constructor) : What class should the smart dict consist of?
            args (dict/other) : Arguments that should be passed to contained_cls. If not a dict, assumed to be the first
                and only argument for the constructor.
            default_args : Default arguments to fill in args

        Returns:
            result (SmartDict instance) : The constructed instance.
        '''

        kwargs = dict_from_defaults_and_variations( default_args, args )

        storage = {}
        for key in kwargs.keys():
            if isinstance( kwargs[key], dict ):
                storage[key] = contained_cls( **kwargs[key] )
            else:
                storage[key] = contained_cls( kwargs[key] )

        return cls( storage )

########################################################################

def merge_two_dicts( dict_a, dict_b ):
    '''Merges two dictionaries into a shallow copy.

    Args:
        dict_a (dict) : First dictionary to merge.
        dict_b (dict) : Second dictionary to merge.

    Returns:
        merged_dict (dict) :
            Dictionary including elements from both.
            dict_a's entries take priority over dict_b.
    '''

    merged_dict = dict_b.copy()
    merged_dict.update( dict_a )

    return merged_dict

########################################################################

def dict_from_defaults_and_variations( defaults, variations ):
    '''Create a dictionary of dictionaries from a default dictionary and variations on it.

    Args:
        defaults (dict) : Default dictionary. What each individual dictionary should default to.
        variations (dict of dicts) : Each dictionary contains what should be different. The key for each dictionary should
            be a label for it.

    Returns:
        result (dict of dicts) : The results is basically variations, where each child dict is merged with defaults.
    '''

    if len( defaults ) == 0:
        return variations

    result = {}
    for key in variations.keys():

        defaults_copy = defaults.copy()

        defaults_copy.update( variations[key] )

        result[key] = defaults_copy

    return result

########################################################################

def deepgetattr( obj, attr ):
    '''Recurses through an attribute chain to get the ultimate value.
    Credit to http://pingfive.typepad.com/blog/2010/04/deep-getattr-python-function.html

    Args:
        obj (object) : Object for which to get the attribute.
        attr (str) : Attribute to get. Can be nested, e.g. obj.foo.bar.dog

    Returns:
        result (attr object) : The requested attribute.
    '''

    return reduce( getattr, attr.split('.'), obj )

########################################################################

def arrays_to_set( *args ):
    '''Convert arr1, arr2, arr3, ... to a set, where a single enrty in the set is ( arr1[j], arr2[j], ... )

    Args:
        n arrays ( multiple np.ndarrays ) : The arrays to convert.

    Returns:
        result ( set ) : The converted set.
    '''

    return set( zip( *args ) )

def set_to_arrays( set_to_convert ):
    '''Convert a set into multiple arrays.

    Args:
        result ( set ) : The set to convert

    Returns:
        n arrays ( multiple np.ndarrays ) : The converted arrays
    '''

    return np.array( list( set_to_convert ) ).transpose()

########################################################################

@contextmanager
def captured_output():
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err

########################################################################

def chunk_list( l, n ):
    '''Breaks a list l into n chunks, as equally as possible.

    Args:
        l (list) : List to break into chunks.
        n (int) : Number of chunks to break the list into.

    Returns:
        chunked_l ( list of lists ) : The list broken into chunks.
    '''

    start_ind = 0
    end_ind = 0
    remainder_to_distribute = len( l ) % n

    chunked_l = []
    for i in range( n ):

        # l/n may not always be an int, so forcing it to be one as before.
        # But the fact that it isn't always an int seems like a bigger problem.
        end_ind += int(len( l )/n)

        if remainder_to_distribute > 0:
            end_ind += 1
            remainder_to_distribute -= 1

        chunked_l.append( l[ start_ind : end_ind ] )

        start_ind = end_ind

    return chunked_l

########################################################################

def get_instance_source_dir( instance, instance_type='class' ):
    '''Get the directory containing the source code of an instance (class or module).

    instance (object) : Instance to get the code for.
    instance_type (str) : 'class' if the instance is a class, 'module' if the instance is a module
    '''

    if instance_type == 'class':
        inspection_object =  instance.__class__
    elif instance_type == 'module':
        inspection_object = instance
    else:
        raise KeyError( "Unrecognized instance_type, {}".format( instance_type ) )

    sourcefile = inspect.getabsfile( inspection_object )

    sourcedir = os.path.dirname( sourcefile )

    return sourcedir

########################################################################

def get_code_version( instance, instance_type='class' ):
    '''Get the current version (git tag/commit) of the parent code for a class or module

    instance (object) : Instance to get the code for.
    instance_type (str) : 'class' if the instance is a class, 'module' if the instance is a module
    '''

    cwd = os.getcwd()

    # Change to the directory of the instance
    instancedir = get_instance_source_dir( instance, instance_type=instance_type )
    os.chdir( instancedir )

    # Get the code version
    try:
        code_version = subprocess.check_output( [ 'git', 'describe', '--always' ] )
    # If we fail, don't break the code.
    except:
        return 'failed to get version'

    # Change back
    os.chdir( cwd )

    return code_version

########################################################################

def print_timer( timer_string='Time taken:' ):
    '''Decorator to time a function.

    Args:
        timer_string (str, optional) : Printed before printing the time.
    '''

    def _print_timer( func ):

        @wraps( func )
        def wrapped_func( *args, **kwargs ):

            time_start = time.time()

            result = func( *args, **kwargs )

            time_end = time.time()

            print_string = '{} {:.3g} seconds'.format( timer_string, time_end - time_start )
            print( print_string )

            return result

        return wrapped_func

    return _print_timer

########################################################################

def store_parameters( constructor ):
    '''Decorator for automatically storing arguments passed to a constructor.
    I.e. any args passed to constructor via test_object = TestObject( *args, **kwargs )
    will be stored in test_object, e.g. test_object.args

    Args:
        constructor (function) : Constructor to wrap.
    '''

    @wraps( constructor )
    def wrapped_constructor( self, *args, **kwargs ):

        parameters_to_store = inspect.getcallargs( constructor, self, *args, **kwargs )

        # Make sure we don't accidentally try to save the self argument
        del parameters_to_store['self']

        for parameter in parameters_to_store.keys():
            setattr( self, parameter, parameters_to_store[parameter] )

        self.stored_parameters = parameters_to_store.keys()

        result = constructor( self, *args, **kwargs )

    return wrapped_constructor

########################################################################

def save_parameters( instance, f ):
    '''Save parameters to a hdf5 file.

    Args:
        instance (object) :
            Instance that has the attributes that are parameters.
            Typically stored in instance.stored_parameters.

        f (open h5py file object) :
            File to save the parameters to.

    Returns:
        param_grp (h5py group) :
            Group containing the parameters as attributes.
    '''

    param_grp = f.create_group( 'parameters' )
    for parameter_str in instance.stored_parameters:

        parameter = getattr( instance, parameter_str )

        try:
            if parameter is None:
                param_grp.attrs[parameter_str] = 'None'
            else:
                param_grp.attrs[parameter_str] = parameter

        except TypeError:
            raise TypeError(
                "Parameter {} = {} failed to save.".format(
                    parameter_str,
                    parameter
                    )
                )

    return param_grp

########################################################################

def gen_derived_data_doc(
    object_import_path,
    save_filepath,
    header_file = None,
):
    '''Generate a document that lists the methods for calculating derived data
    quantities.

    Args:
        object_import_path (str):
            Name of the object to document the derived quantities of.

        save_filepath (str):
            Where to save the result.

        header_file (str):
            What to use as the header for the file, that explains what the
            document contains.
            Defaults to galaxy_dive/utils/derived_data_header.rst.
        
    Modifies:
        file at save_filepath (.rst file):
            Saves the output here.
    '''

    # Process the path
    module_path, class_name = os.path.splitext( object_import_path )
    # Remove the period
    class_name = class_name[1:]

    # Get the object out
    module = importlib.import_module( module_path )
    cl = getattr( module, class_name )

    # Setup the file
    if header_file is None:
        header_file = os.path.join(
            os.path.dirname( __file__ ),
            'derived_data_header.rst',
        )
    shutil.copy( header_file, save_filepath )

    # Write the automatically generated part
    with open( save_filepath, 'a' ) as f:

        # Class line
        f.write( '\n.. autoclass:: {}\n'.format( object_import_path ) )
        f.write( '    :show-inheritance:\n\n' )

        # Write out derived data methods
        for attr in dir( cl ):
            if attr[:5] == 'calc_':
                f.write( '    .. automethod:: {}\n'.format( attr ) )

########################################################################

# Make a path to a file

def make_dir( path ):

    try:
        os.makedirs(path)
    except OSError as exc: # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else: raise

    return 0

########################################################################

# Class modelled after unix tee, mainly for the purpose of writing to a log while also printing.

class Tee:
    def write( self, *args, **kwargs ):
        self.out1.write( *args, **kwargs )
        self.out2.write( *args, **kwargs )
    def __init__( self, out1, out2 ):
        self.out1 = out1
        self.out2 = out2

########################################################################

# Function for saving to a file.
def saveToLog( log_save_file ):

    sys.stdout = Tee(open(log_save_file, 'w'), sys.stdout)

    return 0

########################################################################
# Tools for dealing with cosmology-specific data files.
########################################################################

# Get information from a halo file for a particular redshift.

def getHaloDataRedshift(sdir, halo_number, redshift, convert_to_proper=True):
# Values returned are, in order, and in comoving units,
#   redshift, halo_ID, host_ID, xpeak, ypeak, zpeak, Rvir, Mvir, Mgas, Mstar, xCOM, yCOM, zCOM.

    # Get the halo data for all snapshots
    halo_file = getHaloFilename(sdir, halo_number)
    full_halo_data = dataio.tuples_to_arrays(dataio.read_data(halo_file))
    redshift_arr = full_halo_data[0]

    # Account for rounding errors at the edges.
    if redshift < redshift_arr.min():
        print( 'zhh_dataio.getHaloDataRedshift: Bumping redshift from {} to {}'.format(redshift, redshift_arr.min()) )
        redshift = redshift_arr.min()
    elif redshift > redshift_arr.max():
        print( 'zhh_dataio.getHaloDataRedshift: Bumping redshift from {} to {}'.format(redshift, redshift_arr.max()) )
        redshift = redshift_arr.max()

    try:
        # Get single values out of the halo data.
        arr_value_list = []
        for arr in full_halo_data:
            arr_interp = scipy.interpolate.interp1d(redshift_arr[::-1], arr[::-1])
            arr_value = arr_interp(redshift)
            arr_value_list.append(arr_value)
    except ValueError:
        arr_value_list = np.array(full_halo_data).flatten()

    # Convert halo data to proper coordinates (the 1/h factor's still there for many of the coords)
    if convert_to_proper==True:
        comoving_vals = [3, 4, 5, 6, 10, 11, 12]
        for val_index in comoving_vals:
            arr_value_list[val_index] /= (1. + arr_value_list[0])

    return arr_value_list

########################################################################

# Get particle data and useful halo file data for a given data, all in proper units. Combine them into a class.

class PartDataLong(object):

    def __init__(self, sdir, snum, ptype, halo_number=0):
        self.sdir = sdir
        self.snum = snum
        self.ptype = ptype
        self.halo_number = halo_number

        # Get the particles from a snapshot (in proper units)
        self.P = readsnap.readsnap(sdir, snum, ptype, cosmological=1)

        # Get the halo data
        halo_data = getHaloDataRedshift(sdir, halo_number, self.P['redshift'])

        # Get rid of 1/h factors in the halo data
        vals_to_be_converted = range(3, 13)
        for val_index in vals_to_be_converted:
            halo_data[val_index] /= self.P['hubble']

        # Add the halo data to the class.
        self.redshift = halo_data[0]
        self.halo_ID = halo_data[1]
        self.host_ID = halo_data[2]
        self.peak_coords = (halo_data[3], halo_data[4], halo_data[5])
        self.R_vir = halo_data[6]
        self.M_vir = halo_data[7]
        self.M_gas = halo_data[8]
        self.M_star = halo_data[9]
        self.CoM_coords = (halo_data[10], halo_data[11], halo_data[12])

    ########################################################################

    # Change coordinates to center on the main halo.
    def changeCoordsCenter(self):
        self.P['p'] -= np.array(self.peak_coords)

    ########################################################################

    # Calculate the distance from the origin for a given particle.
    def calcRadialDistance(self):

        coords = self.P['p']
        x, y, z = coords.transpose()
        self.P['r'] = np.sqrt(x**2. + y**2. + z**2.)

        return self.P['r']




#!/usr/bin/env python
'''Tools for plotting AHF data.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import copy
import numpy as np

import matplotlib
matplotlib.use('PDF')
import matplotlib.pyplot as plt
import matplotlib.patheffects as path_effects
import matplotlib.patches as mpatches
import matplotlib.transforms as transforms

import galaxy_dive.utils.astro as astro_tools
import galaxy_dive.plot_data.generic_plotter as generic_plotter

########################################################################

# For catching default values
default = object()

########################################################################
########################################################################


class HaloPlotter( generic_plotter.GenericPlotter ):

    ########################################################################

    def plot_halos_snapshot(
        self,
        snum,
        ax = default,
        type_of_halo_id = 'ahf_halos',
        color = 'w',
        linestyle = 'solid',
        linewidth = 2,
        outline = True,
        center = True,
        hubble_param = default,
        radius_fraction = 1.,
        length_scale = 'Rstar0.5',
        n_halos_plotted = 100,
        show_valid_halos = True,
        minimum_criteria = 'n_star',
        minimum_value = 10,
        show_halo_id = False,
        x_range = None,
        y_range = None,
    ):
        '''Plot the halos as circles at their respective locations.

        Args:
            snum (int) :
                Snapshot to plot at.

            type_of_halo_id (str):
                Should the merger tree halos be plotted, or the first 50 ahf halos at that redshift?

            ax (axis object) :
                Axis to use. If default, one is created.

            color (str) :
                What color should the circle be?

            linestyle (str) :
                What linestyle to use.

            linewidth (int) :
                What linewidth to use.

            outline (bool) :
                Should the circles be outlined for increased visibility?

            center (bool) :
                Should the plot be centered at the most massive halo at z=0?

            hubble_param (float) :
                If given, the positions will be converted to physical kpc.

            radius_fraction (float) :
                The circles will be radius_fraction*r_scale

            length_scale (str) :
                What length scale to use?

            n_halos_plotted (int) :
                Number of AHF halos to show.

            show_valid_halos, minimum_criteria, minimum_value (bool,str,int) :
                If valid_halos is True, only show halos with minimum criteria > minimum_value
        '''

        if ax is default:
            plt.figure( figsize=(7, 6), facecolor='white' )
            ax = plt.gca()

        # Account for the fact that this plotting function was written prior
        # to how we usually get data.
        ahf_reader = self.data_object.ahf_reader

        if type_of_halo_id == 'ahf_halos':

            if show_valid_halos:
                min_criteria = self.data_object.get_data(
                    minimum_criteria, snum )
                has_minimum_value = min_criteria >= minimum_value
                sl = has_minimum_value
            else:
                sl = slice( 0, n_halos_plotted )

            x_pos = self.data_object.get_data( 'Xc', snum )[sl]
            y_pos = self.data_object.get_data( 'Yc', snum )[sl]

            halo_ids = self.data_object.ahf_reader.ahf_halos.index[sl]

            if length_scale == 'r_scale':
                r_vir = ahf_reader.ahf_halos['Rvir'][sl]
                c_analytic = ahf_reader.ahf_halos['cAnalytic'][sl]
                used_length_scale = r_vir / c_analytic
            else:
                used_length_scale = ahf_reader.ahf_halos[length_scale][sl]

        elif type_of_halo_id == 'merger_tree':
            x_pos = ahf_reader.get_mtree_halo_quantity( 'Xc', snum, ahf_reader.index, ahf_reader.tag )
            y_pos = ahf_reader.get_mtree_halo_quantity( 'Yc', snum, ahf_reader.index, ahf_reader.tag )

            if length_scale == 'r_scale':
                r_vir = ahf_reader.get_mtree_halo_quantity( 'Rvir', snum, ahf_reader.index, ahf_reader.tag )
                c_analytic = ahf_reader.get_mtree_halo_quantity(
                    'cAnalytic',
                    snum,
                    ahf_reader.index,
                    ahf_reader.tag
                )
                used_length_scale = r_vir / c_analytic
            else:
                used_length_scale = ahf_reader.get_mtree_halo_quantity(
                    length_scale,
                    snum,
                    ahf_reader.index,
                    ahf_reader.tag
                )

        radii = radius_fraction * used_length_scale

        if hubble_param is not default:
            redshift = self.data_object.get_mt_data(
                'redshift',
                snums = [ snum, ],
            )
            x_pos /= ( 1. + redshift ) * hubble_param
            y_pos /= ( 1. + redshift ) * hubble_param
            radii /= ( 1. + redshift ) * hubble_param

        if center:
            x_center = ahf_reader.mtree_halos[0]['Xc'][snum]
            y_center = ahf_reader.mtree_halos[0]['Yc'][snum]

            if hubble_param is not default:
                x_center /= ( 1. + redshift ) * hubble_param
                y_center /= ( 1. + redshift ) * hubble_param

            x_pos -= x_center
            y_pos -= y_center

        # To set the window size and such automatically, this is easiest
        ax.scatter( x_pos, y_pos, color=color, s=0 )

        for i, radius in enumerate( radii ):

            cir = mpatches.Circle(
                (x_pos[i], y_pos[i]),
                radius=radius,
                linewidth=linewidth,
                color=color,
                linestyle=linestyle,
                fill=False,
                facecolor='w'
            )
            ax.add_patch( cir )

            if outline:
                cir.set_path_effects([
                    path_effects.Stroke(linewidth=5, foreground='black'),
                    path_effects.Normal()
                ])

            if show_halo_id:

                if x_range is not None:
                    if x_pos.values[i] > x_range[1] or x_pos.values[i] < x_range[0]:
                        continue
                if y_range is not None:
                    if y_pos.values[i] > y_range[1] or y_pos.values[i] < y_range[0]:
                        continue

                id_str = '{}'.format( halo_ids.values[i] )
                ax.annotate( s=id_str, xy=(x_pos.values[i], y_pos.values[i]), xycoords='data', )

        if x_range is not None:
            ax.set_xlim( x_range )
        if y_range is not None:
            ax.set_ylim( y_range )

    ########################################################################

    def plot_halos_time(
        self,
        y_key,
        snums,
        n_halos = default,
        subtract_mt_halo_id = None,
        convert_kpc_to_pkpc = True,
        hubble_param = None,
        omega_matter = None,
        ax = None,
        color = 'k',
    ):

        if ax is None:
            plt.figure( figsize=(10, 6), facecolor='white' )
            ax = plt.gca()

        redshifts = self.data_object.get_mt_data( 'redshift', snums=snums )

        x_data = astro_tools.age_of_universe(
            redshifts,
            h = hubble_param,
            omega_matter = omega_matter,
        )

        if subtract_mt_halo_id is not None:
            subtract_data = self.data_object.get_mt_data(
                y_key,
                snums = snums,
                mt_halo_id = subtract_mt_halo_id,
            )

        for i, snum in enumerate( snums ):

            y_data = self.data_object.get_data(
                y_key,
                snum,
            )

            # Cut down on the number of halos plotted
            if n_halos is not default:
                y_data = y_data[:n_halos]

            if subtract_mt_halo_id is not None:
                y_data -= subtract_data[i]

            if convert_kpc_to_pkpc:
                redshift = redshifts[i]
                y_data /= ( 1. + redshift ) * hubble_param

            x_data_to_plot = [ x_data[i], ] * y_data.size

            ax.scatter(
                x_data_to_plot,
                y_data,
                color = 'k',
            )

    ########################################################################

    def plot_mt_halo_time(
        self,
        y_key,
        custom_y_data = None,
        y_scale = default,
        hubble_param = default,
        conversion_factor = default,
        convert_to_physical = False,
        convert_to_comoving = False,
        y_data_div_key = default,
        ax = default,
        halo_ids_to_plot = [ 0, ],
        color = 'k',
        linestyle = '-',
        y_label = default,
        label = default,
        plot_change_in_halo_id = False,
    ):

        if ax is default:
            plt.figure( figsize=(10, 6), facecolor='white' )
            ax = plt.gca()

        for i, halo_id in enumerate( halo_ids_to_plot ):

            plotted_mtree_halo = self.data_object.mt_halos[halo_id]

            x_data = np.log10( 1. + plotted_mtree_halo['redshift'] )

            if custom_y_data is None:
                if y_key == 'r_scale':
                    y_data = plotted_mtree_halo['Rvir'] / \
                        plotted_mtree_halo['cAnalytic']
                else:
                    y_data = plotted_mtree_halo[y_key]
            else:
                y_data = custom_y_data

            # Make a copy of the y-data so we don't alter it.
            y_data = copy.copy( y_data )

            if conversion_factor is not default:
                y_data *= conversion_factor

            if hubble_param is not default:
                y_data /= hubble_param

            if convert_to_physical:
                y_data /= ( 1. + plotted_mtree_halo['redshift'] )

            if convert_to_comoving:
                y_data *= ( 1. + plotted_mtree_halo['redshift'] )

            if y_data_div_key is not default:
                y_data /= plotted_mtree_halo[y_data_div_key]

            # Plot vertical lines when there's a change
            if plot_change_in_halo_id:
                # Make a blended transformation
                trans = transforms.blended_transform_factory( ax.transData, ax.transAxes )

                for i, change_in_halo_id in enumerate( change_in_halo_ids ):
                    if change_in_halo_id != 0:
                        ax.plot(
                            [plotted_mtree_halo.index[i], plotted_mtree_halo.index[i] ],
                            [0., 1.],
                            transform=trans,
                            color='k',
                            linewidth=1,
                            linestyle='--'
                        )

            if label is default:
                label = self.label

            ax.plot( x_data, y_data, color=color, linewidth=3, label=label, linestyle=linestyle )

            if i == 0:
                tick_redshifts = np.array( [ 0.25, 0.5, 1, 2, 3, 4, 5, 6, 7, ] )
                x_tick_values = np.log10( 1. + tick_redshifts )
                plt.xticks( x_tick_values, tick_redshifts )

                if y_label is default:
                    y_label = y_key

                if y_scale is not default:
                    ax.set_yscale( y_scale )

                ax.set_xlabel( r'z', fontsize=22, )
                ax.set_ylabel( y_label, fontsize=22, )
